from collections import namedtuple
import logging
import traceback
from enum import Enum

import attr

from .. import repositories, entities
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class PipelineExecutionStatus(str, Enum):
    PENDING = "pending"
    IN_PROGRESS = "in-progress"
    FAILED = "failed"
    SUCCESS = "success"
    QUEUE = "queue"
    TERMINATED = "terminated"
    RERUN = "rerun"


class CycleRerunMethod(str, Enum):
    START_FROM_NODES = 'startFromNodes',
    START_FROM_FAILED_EXECUTIONS = 'startFromFailedExecutions',
    START_FROM_BEGINNING = 'startFromBeginning'


class PipelineExecutionNode:
    def __init__(self, name, node_id, ports, metadata, node_type, namespace, project_id, status):
        self.node_id = node_id
        self.namespace = namespace
        self.node_type = node_type
        self.status = status
        self.ports = ports
        self.metadata = metadata
        self.project_id = project_id
        self.name = name

    @staticmethod
    def from_json(_json: dict):
        ports = [entities.PipelineNodeIO.from_json(_json=i_input) for i_input in _json.get('ports', list())]
        return PipelineExecutionNode(
            node_id=_json.get('id', None),
            namespace=_json.get('namespace', None),
            node_type=_json.get('type', None),
            status=_json.get('status', None),
            ports=ports,
            metadata=_json.get('metadata', None),
            project_id=_json.get('projectId', None),
            name=_json.get('name', None),
        )

    def to_json(self):
        _json = {
            'id': self.node_id,
            'namespace': self.namespace,
            'type': self.node_type,
            'status': self.status,
            'ports': [_io.to_json() for _io in self.ports],
            'metadata': self.metadata,
            'projectId': self.project_id,
            'name': self.name,
        }

        return _json


@attr.s
class PipelineExecution(entities.BaseEntity):
    """
    Package object
    """
    # platform
    id = attr.ib()
    nodes = attr.ib(repr=False)
    executions = attr.ib(repr=False)
    status = attr.ib()
    # name change
    created_at = attr.ib()
    updated_at = attr.ib(repr=False)
    pipeline_id = attr.ib()
    max_attempts = attr.ib()
    creator = attr.ib()

    # sdk
    _pipeline = attr.ib(repr=False)
    _project = attr.ib(repr=False)
    _client_api = attr.ib(type=ApiClient, repr=False)
    _repositories = attr.ib(repr=False)

    @staticmethod
    def _protected_from_json(_json, client_api, pipeline=None, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error
        :param _json: platform json
        :param client_api: ApiClient entity
        :param pipeline: Pipeline entity
        :param is_fetched: is Entity fetched from Platform
        :return:
        """
        try:
            pipeline = PipelineExecution.from_json(
                _json=_json,
                client_api=client_api,
                pipeline=pipeline,
                is_fetched=is_fetched
            )
            status = True
        except Exception:
            pipeline = traceback.format_exc()
            status = False
        return status, pipeline

    @classmethod
    def from_json(cls, _json, client_api, pipeline=None, is_fetched=True) -> 'PipelineExecution':
        """
        Turn platform representation of pipeline_execution into a pipeline_execution entity

        :param dict _json: platform representation of package
        :param dl.ApiClient client_api: ApiClient entity
        :param dtlpy.entities.pipeline.Pipeline pipeline: Pipeline entity
        :param bool is_fetched: is Entity fetched from Platform
        :return: Pipeline entity
        :rtype: dtlpy.entities.PipelineExecution
        """
        project = None
        if pipeline is not None:
            project = pipeline._project
            if pipeline.id != _json.get('pipelineId', None):
                logger.warning('Pipeline has been fetched from a project that is not belong to it')
                pipeline = None

        nodes = [PipelineExecutionNode.from_json(_json=node) for node in _json.get('nodes', list())]

        inst = cls(
            id=_json.get('id', None),
            created_at=_json.get('createdAt', None),
            updated_at=_json.get('updatedAt', None),
            pipeline_id=_json.get('pipelineId', None),
            status=_json.get('status', None),
            max_attempts=_json.get('maxAttempts', None),
            creator=_json.get('creator', None),
            nodes=nodes,
            executions=_json.get('executions', dict()),
            pipeline=pipeline,
            project=project,
            client_api=client_api,
        )

        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Turn Package entity into a platform representation of Package

        :return: platform json of package
        :rtype: dict
        """
        _json = attr.asdict(self,
                            filter=attr.filters.exclude(attr.fields(PipelineExecution)._repositories,
                                                        attr.fields(PipelineExecution)._client_api,
                                                        attr.fields(PipelineExecution)._pipeline,
                                                        attr.fields(PipelineExecution).nodes,
                                                        attr.fields(PipelineExecution).created_at,
                                                        attr.fields(PipelineExecution).updated_at,
                                                        attr.fields(PipelineExecution).pipeline_id,
                                                        attr.fields(PipelineExecution).executions,
                                                        attr.fields(PipelineExecution).max_attempts
                                                        ))
        executions = dict()
        for node_id, executions_list in self.executions.items():
            if len(executions_list) > 0 and isinstance(executions_list[0], entities.Execution):
                executions[node_id] = [e.to_json() for e in executions_list]
            else:
                executions[node_id] = executions_list

        _json['pipelineId'] = self.pipeline_id
        _json['maxAttempts'] = self.max_attempts
        _json['createdAt'] = self.created_at
        _json['updatedAt'] = self.updated_at
        _json['nodes'] = [node.to_json() for node in self.nodes]
        _json['executions'] = executions
        return _json

    #########
    # Props #
    #########
    @property
    def pipeline(self):
        if self._pipeline is None:
            self._pipeline = self.pipelines.get(pipeline_id=self.pipeline_id, fetch=None)
        assert isinstance(self._pipeline, entities.Pipeline)
        return self._pipeline

    @property
    def project(self):
        if self._project is None:
            self._project = self.pipeline.project
        assert isinstance(self._pipeline.project, entities.Project)
        return self._pipeline.project

    ################
    # repositories #
    ################
    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['projects', 'pipelines', 'pipeline_executions'])

        r = reps(
            projects=repositories.Projects(client_api=self._client_api),
            pipelines=repositories.Pipelines(client_api=self._client_api, project=self._project),
            pipeline_executions=repositories.PipelineExecutions(client_api=self._client_api,
                                                                project=self._project,
                                                                pipeline=self._pipeline)
        )
        return r

    @property
    def projects(self):
        assert isinstance(self._repositories.projects, repositories.Projects)
        return self._repositories.projects

    @property
    def pipelines(self):
        assert isinstance(self._repositories.pipelines, repositories.Pipelines)
        return self._repositories.pipelines

    @property
    def pipeline_executions(self):
        assert isinstance(self._repositories.pipeline_executions, repositories.PipelineExecutions)
        return self._repositories.pipeline_executions

    def rerun(self,
              method: str = None,
              start_nodes_ids: list = None,
              wait: bool = True
              ) -> bool:
        """
        Get Pipeline Execution object

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param str method: method to run
        :param list start_nodes_ids: list of start nodes ids
        :param bool wait: wait until rerun finish
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            pipeline_executions.rerun(method=dl.CycleRerunMethod.START_FROM_BEGINNING,)
        """
        filters = entities.Filters(field='id', values=[self.id], operator=entities.FiltersOperations.IN,
                                   resource=entities.FiltersResource.PIPELINE_EXECUTION)
        return self.pipeline_executions.rerun(
            method=method,
            start_nodes_ids=start_nodes_ids,
            filters=filters,
            wait=wait
        )

    def wait(self):
        """
        Wait for pipeline execution

        :return: Pipeline execution object
        """
        return self.pipeline_executions.wait(pipeline_execution_id=self.id)

    def in_progress(self):
        return self.status not in [PipelineExecutionStatus.FAILED,
                                   PipelineExecutionStatus.SUCCESS,
                                   PipelineExecutionStatus.TERMINATED]



import dtlpy as dl
import logging

logger = logging.getLogger(name='dtlpy')


class ServiceRunner(dl.BaseServiceRunner):
    """
    Package runner class

    """

    def __init__(self):
        """
        Init package attributes here

        :return:
        """



#! /usr/bin/env python3
# This file is part of DTLPY.
#
# DTLPY is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DTLPY is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with DTLPY.  If not, see <http://www.gnu.org/licenses/>.
from . import package_defaults
from .annotation_definitions import BaseAnnotationDefinition
from .base_entity import BaseEntity, DlEntity, DlProperty, DlList, EntityScopeLevel
from .item import Item, ItemStatus, Modality, ModalityTypeEnum, ModalityRefTypeEnum, ExportMetadata
from .links import Link, ItemLink, UrlLink, LinkTypeEnum
from .trigger import Trigger, TriggerResource, TriggerAction, TriggerExecutionMode, BaseTrigger, CronTrigger, \
    TriggerType
from .project import Project, MemberRole
from .artifact import ItemArtifact, LocalArtifact, LinkArtifact, ArtifactType, Artifact
from .dataset import Dataset, ExpirationOptions, IndexDriver, ExportType, OutputExportType
from .codebase import Codebase
from .annotation import Annotation, FrameAnnotation, ViewAnnotationOptions, AnnotationStatus, AnnotationType, \
    ExportVersion
from .annotation_collection import AnnotationCollection
from .paged_entities import PagedEntities
from .filters import Filters, FiltersKnownFields, FiltersResource, FiltersOperations, FiltersMethod, \
    FiltersOrderByDirection
from .recipe import Recipe
from .ontology import Ontology, AttributesTypes, AttributesRange
from .annotation_definitions import Box, Cube, Cube3d, Point, Segmentation, Polygon, Ellipse, Classification, \
    Subtitle, Text, FreeText, RefImage, \
    Polyline, Comparison, UndefinedAnnotationType, Note, Message, Description, Pose, Gis, GisType
from .label import Label
from .codebase import Codebase, PackageCodebaseType, ItemCodebase, GitCodebase, FilesystemCodebase, LocalCodebase
from .package import Package, RequirementOperator, PackageRequirement
from .package_module import PackageModule
from .package_slot import PackageSlot, SlotPostAction, SlotPostActionType, SlotDisplayScope, SlotDisplayScopeResource, \
    UiBindingPanel
from .package_function import PackageFunction, FunctionIO, PackageInputType
from .time_series import TimeSeries
from .service import Service, KubernetesAutoscalerType, KubernetesRabbitmqAutoscaler, KubernetesAutoscaler, KubernetesRPSAutoscaler, \
    InstanceCatalog, KubernetesRuntime, ServiceType, ServiceModeType, DynamicConcurrencyUpdateMethod
from .execution import Execution, ExecutionStatus
from .command import Command, CommandsStatus
from .assignment import Assignment, Workload, WorkloadUnit
from .task import Task, ItemAction, TaskPriority, ConsensusTaskType, AllocationMethod
from .directory_tree import DirectoryTree
from .user import User
from .bot import Bot
from .webhook import Webhook, HttpMethod
from .model import Model, DatasetSubsetType, PlotSample, ModelStatus
from .driver import Driver, S3Driver, GcsDriver, AzureBlobDriver
from .pipeline import Pipeline, PipelineStats, PipelineResumeOption, PipelineSettings, Variable, CompositionStatus
from .node import PipelineConnection, PipelineNode, PipelineConnectionPort, PipelineNodeIO, TaskNode, \
    CodeNode, FunctionNode, PipelineNodeType, PipelineNameSpace, DatasetNode
from .pipeline_execution import PipelineExecution, PipelineExecutionNode, PipelineExecutionStatus, CycleRerunMethod
from .feature import Feature
from .feature_set import FeatureSet, FeatureEntityType
from .organization import Organization, OrganizationsPlans, MemberOrgRole, CacheAction, PodType
from .analytic import ServiceSample, ExecutionSample, PipelineExecutionSample
from .integration import Integration, IntegrationType
from .driver import Driver, ExternalStorage
from .setting import Role, PlatformEntityType, SettingsValueTypes, SettingsTypes, SettingsSectionNames, SettingScope, \
    BaseSetting, UserSetting, Setting
from .reflect_dict import ReflectDict
from .dpk import Dpk, Panel, Toolbar, Components, DpkComputeConfig, DpkComponentChannel, PipelineNode, \
    CustomNodeScope, ToolbarInvoke
from .app import App, AppScope
from .app_module import AppModule
from .resource_execution import ResourceExecution
from .message import Message, NotificationEventContext
from .prompt_item import Prompt, PromptItem, PromptType
from .compute import ClusterProvider, ComputeType, ComputeStatus, Toleration, DeploymentResource, DeploymentResources, ComputeSettings, ComputeConsumptionMethod, \
    NodePool, AuthenticationIntegration, Authentication, ComputeCluster, ComputeContext, Compute, KubernetesCompute
from .service_driver import ServiceDriver
from .gis_item import  ItemGis, Layer
from .collection import Collection




import argparse


def get_parser():
    """
    Build the parser for CLI
    :return: parser object
    """
    parser = argparse.ArgumentParser(
        description="CLI for Dataloop",
        formatter_class=argparse.RawTextHelpFormatter
    )

    ###############
    # sub parsers #
    ###############
    subparsers = parser.add_subparsers(dest="operation", help="supported operations")

    ########
    # shell #
    ########
    subparsers.add_parser("shell", help="Open interactive Dataloop shell")

    ########
    # shell #
    ########
    a = subparsers.add_parser("upgrade", help="Update dtlpy package")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-u", "--url", metavar='\b', help="Package url. default 'dtlpy'", default=None)

    ##################
    # Login / Logout #
    ##################
    subparsers.add_parser("logout", help="Logout")

    subparsers.add_parser("login", help="Login using web Auth0 interface")

    a = subparsers.add_parser("login-token", help="Login by passing a valid token")
    required = a.add_argument_group("required named arguments")
    required.add_argument(
        "-t", "--token", metavar='\b', help="valid token", required=True
    )

    a = subparsers.add_parser("login-secret", help="Login client id and secret")
    required = a.add_argument_group("required named arguments")
    required.add_argument(
        "-e", "--email", metavar='\b', help="user email", required=False, default=None
    )
    required.add_argument(
        "-p", "--password", metavar='\b', help="user password", required=False, default=None
    )
    required.add_argument(
        "-i", "--client-id", metavar='\b', help="client id", required=False, default=None
    )
    required.add_argument(
        "-s", "--client-secret", metavar='\b', help="client secret", required=False, default=None
    )

    a = subparsers.add_parser("login-m2m", help="Login client id and secret")
    required = a.add_argument_group("required named arguments")
    required.add_argument(
        "-e", "--email", metavar='\b', help="user email", required=False, default=None
    )
    required.add_argument(
        "-p", "--password", metavar='\b', help="user password", required=False, default=None
    )
    required.add_argument(
        "-i", "--client-id", metavar='\b', help="client id", required=False, default=None
    )
    required.add_argument(
        "-s", "--client-secret", metavar='\b', help="client secret", required=False, default=None
    )
    ########
    # Init #
    ########
    subparsers.add_parser("init", help="Initialize a .dataloop context")

    ##################
    # Checkout state #
    ##################
    subparsers.add_parser("checkout-state", help="Print checkout state")

    ########
    # Help #
    ########
    subparsers.add_parser("help", help="Get help")

    ###########
    # version #
    ###########
    parser.add_argument("-v", "--version", action="store_true", help="dtlpy version")
    subparsers.add_parser("version", help="DTLPY SDK version")

    #######
    # API #
    #######
    subparser = subparsers.add_parser("api", help="Connection and environment")
    subparser_parser = subparser.add_subparsers(dest="api", help="gate operations")

    # ACTIONS #

    # info
    subparser_parser.add_parser("info", help="Print api information")

    # setenv
    a = subparser_parser.add_parser("setenv", help="Set platform environment")
    required = a.add_argument_group("required named arguments")
    required.add_argument("-e", "--env", metavar='\b', help="working environment", required=True)

    ###############
    # Development #
    ###############
    subparser = subparsers.add_parser(name="development", help="Start a development session")
    subparser_parser = subparser.add_subparsers(dest="development", help="local development session")
    local_subparser_parser = subparser_parser.add_parser(name="local",
                                                         help="local development using docker and code-server session")
    remote_subparser_parser = subparser_parser.add_parser(name="remote",
                                                          help="remote development (dl.Service) using code-server session")

    #############################
    # Local Development Session #
    #############################
    # start
    local_subparser_subparser = local_subparser_parser.add_subparsers(dest="local", help="Local development session")
    a = local_subparser_subparser.add_parser(name="start", help="Start a local development session")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-p", "--port",
                          default=None,
                          help="Local port for the docker connection, default: 5802")
    optional.add_argument("-d", "--docker-image",
                          default=None,
                          help="Docker image to create. default: 'dataloopai/dtlpy-agent:1.57.3.gpu.cuda11.5.py3.8.opencv'")
    # pause
    _ = local_subparser_subparser.add_parser(name="pause", help="Pause the local development session (container pause)")
    # stop
    _ = local_subparser_subparser.add_parser(name="stop", help="Stop the local development session (container kill)")
    ##############################
    # Remote Development Session #
    ##############################

    ############
    # Projects #
    ############
    subparser = subparsers.add_parser("projects", help="Operations with projects")
    subparser_parser = subparser.add_subparsers(dest="projects", help="projects operations")

    # ACTIONS #

    # list
    subparser_parser.add_parser("ls", help="List all projects")

    # create
    a = subparser_parser.add_parser("create", help="Create a new project")
    required = a.add_argument_group("required named arguments")
    required.add_argument("-p", "--project-name", metavar='\b', help="project name")

    # checkout
    a = subparser_parser.add_parser("checkout", help="checkout a project")
    required = a.add_argument_group("required named arguments")
    required.add_argument("-p", "--project-name", metavar='\b', help="project name")

    # open web
    a = subparser_parser.add_parser("web", help="Open in web browser")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-p", "--project-name", metavar='\b', help="project name")

    ############
    # Datasets #
    ############
    subparser = subparsers.add_parser("datasets", help="Operations with datasets")
    subparser_parser = subparser.add_subparsers(dest="datasets", help="datasets operations")

    # ACTIONS #
    # open web
    a = subparser_parser.add_parser("web", help="Open in web browser")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-p", "--project-name", metavar='\b', help="project name")
    optional.add_argument("-d", "--dataset-name", metavar='\b', help="dataset name")

    # list
    a = subparser_parser.add_parser("ls", help="List of datasets in project")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-p", "--project-name", metavar='\b', default=None,
                          help="project name. Default taken from checked out (if checked out)")

    # create
    a = subparser_parser.add_parser("create", help="Create a new dataset")
    required = a.add_argument_group("required named arguments")
    required.add_argument("-d", "--dataset-name", metavar='\b', help="dataset name", required=True)
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-p", "--project-name", metavar='\b', default=None,
                          help="project name. Default taken from checked out (if checked out)")
    optional.add_argument("-c", "--checkout", action='store_true', default=False, help="checkout the new dataset")

    # checkout
    a = subparser_parser.add_parser("checkout", help="checkout a dataset")
    required = a.add_argument_group("required named arguments")
    required.add_argument("-d", "--dataset-name", metavar='\b', help="dataset name")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-p", "--project-name", metavar='\b', default=None,
                          help="project name. Default taken from checked out (if checked out)")

    #########
    # items #
    #########
    subparser = subparsers.add_parser("items", help="Operations with items")
    subparser_parser = subparser.add_subparsers(dest="items", help="items operations")

    # ACTIONS #

    a = subparser_parser.add_parser("web", help="Open in web browser")
    required = a.add_argument_group("required named arguments")
    required.add_argument("-r", "--remote-path", metavar='\b', help="remote path")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-p", "--project-name", metavar='\b', help="project name")
    optional.add_argument("-d", "--dataset-name", metavar='\b', help="dataset name")

    # list
    a = subparser_parser.add_parser("ls", help="List of items in dataset")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-p", "--project-name", metavar='\b', default=None,
                          help="project name. Default taken from checked out (if checked out)")
    optional.add_argument("-d", "--dataset-name", metavar='\b', default=None,
                          help="dataset name. Default taken from checked out (if checked out)")
    optional.add_argument("-o", "--page", metavar='\b', help="page number (integer)", default=0)
    optional.add_argument("-r", "--remote-path", metavar='\b', help="remote path", default=None)
    optional.add_argument("-t", "--type", metavar='\b', help="Item type", default=None)

    # upload
    a = subparser_parser.add_parser("upload", help="Upload directory to dataset")
    required = a.add_argument_group("required named arguments")
    required.add_argument("-l", "--local-path", required=True, metavar='\b',
                          help="local path")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-p", "--project-name", metavar='\b', default=None,
                          help="project name. Default taken from checked out (if checked out)")
    optional.add_argument("-d", "--dataset-name", metavar='\b', default=None,
                          help="dataset name. Default taken from checked out (if checked out)")
    optional.add_argument("-r", "--remote-path", metavar='\b', default=None,
                          help="remote path to upload to. default: /")
    optional.add_argument("-f", "--file-types", metavar='\b', default=None,
                          help='Comma separated list of file types to upload, e.g ".jpg,.png". default: all')
    optional.add_argument("-lap", "--local-annotations-path", metavar='\b', default=None,
                          help="Path for local annotations to upload with items")
    optional.add_argument("-ow", "--overwrite", dest="overwrite", action='store_true', default=False,
                          help="Overwrite existing item")

    # download
    a = subparser_parser.add_parser("download", help="Download dataset to a local directory")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-p", "--project-name", metavar='\b', default=None,
                          help="project name. Default taken from checked out (if checked out)")
    optional.add_argument("-d", "--dataset-name", metavar='\b', default=None,
                          help="dataset name. Default taken from checked out (if checked out)")
    optional.add_argument("-ao", "--annotation-options", metavar='\b',
                          help="which annotation to download. options: json,instance,mask", default=None)
    optional.add_argument("-aft", "--annotation-filter-type", metavar='\b',
                          help="annotation type filter when downloading annotations. "
                               "options: box,segment,binary etc", default=None)
    optional.add_argument("-afl", "--annotation-filter-label", metavar='\b',
                          help="labels filter when downloading annotations.", default=None)
    optional.add_argument("-r", "--remote-path", metavar='\b', default=None,
                          help="remote path to upload to. default: /")
    optional.add_argument("-ow", "--overwrite", action='store_true', default=False,
                          help="Overwrite existing item")
    optional.add_argument("-t", "--not-items-folder", action='store_true', default=False,
                          help="Download WITHOUT 'items' folder")
    optional.add_argument("-wt", "--with-text", action='store_true', default=False,
                          help="Annotations will have text in mask")
    optional.add_argument("-th", "--thickness", metavar='\b', default="1",
                          help="Annotation line thickness")
    optional.add_argument("-l", "--local-path", metavar='\b', default=None,
                          help="local path")
    optional.add_argument("-wb", "--without-binaries", action='store_true', default=False,
                          help="Don't download item binaries")

    ##########
    # videos #
    ##########
    subparser = subparsers.add_parser("videos", help="Operations with videos")
    subparser_parser = subparser.add_subparsers(dest="videos", help="videos operations")

    # ACTIONS #

    # play
    a = subparser_parser.add_parser("play", help="Play video")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument(
        "-l",
        "--item-path",
        metavar='\b',
        default=None,
        help="Video remote path in platform. e.g /dogs/dog.mp4",
    )
    optional.add_argument(
        "-p",
        "--project-name",
        metavar='\b',
        default=None,
        help="project name. Default taken from checked out (if checked out)",
    )
    optional.add_argument(
        "-d",
        "--dataset-name",
        metavar='\b',
        default=None,
        help="dataset name. Default taken from checked out (if checked out)",
    )

    # upload
    a = subparser_parser.add_parser("upload", help="Upload a single video")
    required = a.add_argument_group("required named arguments")
    required.add_argument(
        "-f", "--filename", metavar='\b', help="local filename to upload", required=True
    )
    required.add_argument(
        "-p", "--project-name", metavar='\b', help="project name", required=True
    )
    required.add_argument(
        "-d", "--dataset-name", metavar='\b', help="dataset name", required=True
    )
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument(
        "-r", "--remote-path", metavar='\b', help="remote path", default="/"
    )

    # split video to chunks
    optional.add_argument(
        "-sc",
        "--split-chunks",
        metavar='\b',
        default=None,
        help="Video splitting parameter: Number of chunks to split",
    )
    optional.add_argument(
        "-ss",
        "--split-seconds",
        metavar='\b',
        default=None,
        help="Video splitting parameter: Seconds of each chuck",
    )
    optional.add_argument(
        "-st",
        "--split-times",
        metavar='\b',
        default=None,
        help="Video splitting parameter: List of seconds to split at. e.g 600,1800,2000",
    )
    # encode
    optional.add_argument(
        "-e",
        "--encode",
        action="store_true",
        default=False,
        help="encode video to mp4, remove bframes and upload",
    )

    ###############
    # Application #
    ###############
    subparser = subparsers.add_parser("app", help="Operations with application")
    subparser_parser = subparser.add_subparsers(dest="app", help="application operations")

    # ACTIONS #

    # init
    a = subparser_parser.add_parser('init', help="Initialize the structure in order to deploy a dpk")
    optional = a.add_argument_group("Optional named arguments")
    optional.add_argument('--name', required=False, dest='name', help="the name of the app")
    optional.add_argument('--description', required=False, dest='description', help="the description of the app")
    optional.add_argument('--attributes', required=False, dest='attributes',
                          help="the attributes of the app (comma separated)")
    optional.add_argument('--icon', required=False, dest='icon', help="the icon of the app")
    optional.add_argument('--scope', required=False, dest='scope',
                          help="the scope of the app (default is organization)")
    # add components
    a = subparser_parser.add_parser('add', help="Add component to the dataloop json file")
    optional = a.add_argument_group("Optional named arguments")
    optional.add_argument('--panel', action='store_true', help="builder for panel component")
    optional.add_argument('--module', action='store_true', help="builder for module component")
    optional.add_argument('--toolbar', action='store_true', help="builder for toolbar component")

    # pack
    a = subparser_parser.add_parser("pack", help="Pack the project as dpk file")
    # publish
    a = subparser_parser.add_parser("publish", help="Publish the app")
    optional = a.add_argument_group("Optional named arguments")
    optional.add_argument("--project-name", dest="project_name", help="The name of the project")
    optional.add_argument("--project-id", dest="project_id", help="The ID of the project")

    # update
    a = subparser_parser.add_parser("update", help="Update the app")
    required = a.add_argument_group("Required named arguments")
    required.add_argument("--app-name", dest="app_name", required=True, help="Locates the app by the name")
    required.add_argument("--new-version", dest="new_version", required=True,
                          help="Sets the new version of the specified app")
    optional = a.add_argument_group("Required named arguments")
    optional.add_argument("--project-id", dest="project_id", default=None, help="The id of the project")
    optional.add_argument("--project-name", dest="project_name", default=None, help="The name of the project")

    # install
    a = subparser_parser.add_parser("install", help="Install the app to the platform")
    required = a.add_argument_group("Required named arguments")
    required.add_argument("--dpk-id", dest="dpk_id", required=True, help="The id of the dpk")
    optional = a.add_argument_group("Optional named arguments")
    optional.add_argument("--project-id", dest="project_id", default=None, help="The id of the project")
    optional.add_argument("--project-name", dest="project_name", default=None, help="The name of the project")
    optional.add_argument("--org-id", dest="org_id", default=None, help="The name of the org")

    # pull
    a = subparser_parser.add_parser('pull', help="Pull the app from the marketplace")
    required = a.add_argument_group("Required named arguments")
    required.add_argument('--dpk-name', dest='app_name', required=True, help='The name of the dpk')

    # list
    a = subparser_parser.add_parser('list', help="List all installed apps")
    required = a.add_argument_group("Required named arguments")
    required.add_argument('--project-name', dest='project_name', required=False, help='The name of the project')

    ############
    # Services #
    ############
    subparser = subparsers.add_parser("services", help="Operations with services")
    subparser_parser = subparser.add_subparsers(dest="services", help="services operations")

    # ACTIONS #

    # execute
    a = subparser_parser.add_parser("execute", help="Create an execution")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-f", "--function-name", dest="function_name", default=None,
                          help="which function to run")
    optional.add_argument("-s", "--service-name", dest="service_name", default=None,
                          help="which service to run")
    optional.add_argument("-pr", "--project-name", dest="project_name", default=None,
                          help="Project name")
    optional.add_argument("-as", "--async", dest="asynchronous", default=True, action='store_false',
                          help="Async execution ")
    optional.add_argument("-i", "--item-id", dest="item_id", default=None,
                          help="Item input")
    optional.add_argument("-d", "--dataset-id", dest="dataset_id", default=None,
                          help="Dataset input")
    optional.add_argument("-a", "--annotation-id", dest="annotation_id", default=None,
                          help="Annotation input")
    optional.add_argument("-in", "--inputs", dest="inputs", default='{}',
                          help="Dictionary string input")

    # tear-down
    a = subparser_parser.add_parser(
        "tear-down", help="tear-down service of service.json file"
    )
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-l", "--local-path", dest="local_path", default=None,
                          help="path to service.json file")
    optional.add_argument("-pr", "--project-name", dest="project_name", default=None,
                          help="Project name")

    # ls
    a = subparser_parser.add_parser("ls", help="List project's services")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-pr", "--project-name", dest="project_name", default=None,
                          help="Project name")
    optional.add_argument("-pkg", "--package-name", dest="package_name", default=None,
                          help="Package name")

    # log
    a = subparser_parser.add_parser("log", help="Get services log")
    optional = a.add_argument_group("required named arguments")
    optional.add_argument("-pr", "--project-name", dest="project_name", default=None,
                          help="Project name")
    optional.add_argument("-f", "--service-name", dest="service_name", default=None,
                          help="Project name")
    optional.add_argument("-t", "--start", dest="start", default=None,
                          help="Log start time")

    # delete
    a = subparser_parser.add_parser("delete", help="Delete Service")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-f", "--service-name", dest="service_name", default=None,
                          help="Service name")
    optional.add_argument("-p", "--project-name", dest="project_name", default=None,
                          help="Project name")
    optional.add_argument("-pkg", "--package-name", dest="package_name", default=None,
                          help="Package name")

    ############
    # Triggers #
    ############
    subparser = subparsers.add_parser("triggers", help="Operations with triggers")
    subparser_parser = subparser.add_subparsers(dest="triggers", help="triggers operations")

    # ACTIONS #
    # create
    a = subparser_parser.add_parser("create", help="Create a Service Trigger")
    required = a.add_argument_group("required named arguments")
    required.add_argument("-r", "--resource", dest="resource",
                          help="Resource name", required=True)
    required.add_argument("-a", "--actions", dest="actions", help="Actions", required=True)

    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-p", "--project-name", dest="project_name", default=None,
                          help="Project name")
    optional.add_argument("-pkg", "--package-name", dest="package_name", default=None,
                          help="Package name")
    optional.add_argument("-f", "--service-name", dest="service_name",
                          help="Service name", default=None)
    optional.add_argument("-n", "--name", dest="name",
                          help="Trigger name", default=None)
    optional.add_argument("-fl", "--filters", dest="filters", default='{}',
                          help="Json filter")
    optional.add_argument("-fn", "--function-name", dest="function_name", default='run',
                          help="Function name")

    # delete
    a = subparser_parser.add_parser("delete", help="Delete Trigger")
    required = a.add_argument_group("required named arguments")
    required.add_argument("-t", "--trigger-name", dest="trigger_name", default=None,
                          help="Trigger name", required=True)

    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-f", "--service-name", dest="service_name", default=None,
                          help="Service name")
    optional.add_argument("-p", "--project-name", dest="project_name", default=None,
                          help="Project name")
    optional.add_argument("-pkg", "--package-name", dest="package_name", default=None,
                          help="Package name")

    a = subparser_parser.add_parser("ls", help="List triggers")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-pr", "--project-name", dest="project_name", default=None,
                          help="Project name")
    optional.add_argument("-pkg", "--package-name", dest="package_name", default=None,
                          help="Package name")
    optional.add_argument("-s", "--service-name", dest="service_name", default=None,
                          help="Service name")

    ############
    # Deploy   #
    ############
    # subparsers.add_parser("deploy", help="Login using web Auth0 interface")

    a = subparsers.add_parser("deploy", help="deploy with json file")
    required = a.add_argument_group("required named arguments")
    required.add_argument("-f", dest="json_file", default=None,
                          help="Path to json file")
    required.add_argument("-p", dest="project_name", default=None,
                          help="Project name")

    ############
    # Generate   #
    ############
    # subparsers.add_parser("deploy", help="Login using web Auth0 interface")

    a = subparsers.add_parser("generate", help="generate a json file")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("--option", dest="package_type", default=None,
                          help="cataluge of examples")
    optional.add_argument("-p", "--package-name", dest="package_name", default=None,
                          help="Package name")

    ############
    # packages #
    ############
    subparser = subparsers.add_parser("packages", help="Operations with packages")
    subparser_parser = subparser.add_subparsers(
        dest="packages", help="package operations"
    )

    # ACTIONS #
    # ls
    a = subparser_parser.add_parser("ls", help="List packages")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-p", "--project-name", dest="project_name", default=None,
                          help="Project name")

    # push
    a = subparser_parser.add_parser("push", help="Create package in platform")

    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-src", "--src-path", metavar='\b', default=None,
                          help="Revision to deploy if selected True")
    optional.add_argument("-cid", "--codebase-id", metavar='\b', default=None,
                          help="Revision to deploy if selected True")
    optional.add_argument("-pr", "--project-name", metavar='\b', default=None,
                          help="Project name")
    optional.add_argument("-p", "--package-name", metavar='\b', default=None,
                          help="Package name")
    optional.add_argument("-c", "--checkout", action='store_true', default=False, help="checkout the new package")

    # deploy
    a = subparser_parser.add_parser("deploy", help="Deploy package to platform")
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-p", "--package-name", metavar='\b', default=None,
                          help="Package name")
    optional.add_argument("-pr", "--project-name", metavar='\b', default=None,
                          help="Project name")
    optional.add_argument("--module-name", metavar='\b', default='default_module',
                          help="Package module name")
    optional.add_argument("-c", "--checkout", action='store_true', default=False, help="checkout the new package")

    # test
    a = subparser_parser.add_parser(
        "test", help="Tests that Package locally using mock.json"
    )
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-c", "--concurrency", metavar='\b', default=10,
                          help="Revision to deploy if selected True")
    optional.add_argument("-f", "--function-name", metavar='\b', default='run',
                          help="Function to test")
    # checkout
    a = subparser_parser.add_parser("checkout", help="checkout a package")
    required = a.add_argument_group("required named arguments")
    required.add_argument("-p", "--package-name", metavar='\b', help="package name")

    # delete
    a = subparser_parser.add_parser(
        "delete", help="Delete Package"
    )
    optional = a.add_argument_group("optional named arguments")
    optional.add_argument("-pkg", "--package-name", dest="package_name", default=None,
                          help="Package name")

    optional.add_argument("-p", "--project-name", dest="project_name", default=None,
                          help="Project name")

    #########
    # Shell #
    #########
    # ls
    subparsers.add_parser("ls", help="List directories")
    #
    # pwd
    subparsers.add_parser("pwd", help="Get current working directory")

    # cd
    subparser = subparsers.add_parser("cd", help="Change current working directory")
    subparser.add_argument(dest='dir')

    # mkdir
    subparser = subparsers.add_parser("mkdir", help="Make directory")
    subparser.add_argument(dest='name')

    # clear
    subparsers.add_parser("clear", help="Clear shell")

    ########
    # Exit #
    ########
    subparsers.add_parser("exit", help="Exit interactive shell")

    return parser



#! /usr/bin/env python3
# This file is part of DTLPY.
#
# DTLPY is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DTLPY is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with DTLPY.  If not, see <http://www.gnu.org/licenses/>.
import warnings
import logging
import sys
import os

##########
# Logger #
##########
from .services import DataloopLogger, DtlpyFilter
from .__version__ import version as __version__

logger = logging.getLogger(name='dtlpy')
logging.getLogger(name='filelock').setLevel(level=logging.CRITICAL)
if len(logger.handlers) == 0:
    logger.setLevel(logging.DEBUG)
    log_filepath = DataloopLogger.get_log_filepath()
    # set file handler to save all logs to file
    stream_formatter = logging.Formatter(
        fmt="[%(asctime)-s][%(levelname).3s][%(name)s:v" + __version__ + "][%(relativepath)-s:%(lineno)-d] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    file_formatter = logging.Formatter(
        fmt="[%(asctime)s.%(msecs)03d][%(threadName)s][%(levelname).3s][%(name)s:v" +
            __version__ + "][%(relativepath)-s:%(lineno)-d](%(funcName)-s): %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    package_path = os.path.dirname(__file__)
    # relative function path filtering
    filtering = DtlpyFilter(package_path)
    fh = DataloopLogger(log_filepath, maxBytes=(1048 * 1000 * 5))
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(file_formatter)
    fh.addFilter(filtering)
    sh = logging.StreamHandler()
    sh.setLevel(logging.WARNING)
    sh.setFormatter(stream_formatter)
    sh.addFilter(filtering)
    # set handlers to main logger
    logger.addHandler(sh)
    logger.addHandler(fh)

from .services.api_client import client as client_api
from .services.api_client import VerboseLoggingLevel, ApiClient
from .services import DataloopLogger, DtlpyFilter, check_sdk, Reporter, service_defaults
from .services.api_reference import api_reference as _api_reference
from .caches.cache import CacheConfig, CacheType
from .exceptions import PlatformException
from . import repositories, exceptions, entities, examples
from .entities import (
    # main entities
    Project, Dataset, ExpirationOptions, ExportVersion, Trigger, Item, Execution, AnnotationCollection, Annotation,
    Recipe, IndexDriver, AttributesTypes, AttributesRange, Dpk, App, AppModule, AppScope,
    Ontology, Label, Task, TaskPriority, ConsensusTaskType, AllocationMethod, Assignment, Service, Package, Codebase, Model,
    PackageModule, PackageFunction,
    # annotations
    Box, Cube, Cube3d, Point, Note, Message, Segmentation, Ellipse, Classification, Subtitle, Polyline, Pose, Gis, GisType,
    Description,
    Polygon, Text, FreeText, RefImage,
    # filters
    Filters, FiltersKnownFields, FiltersResource, FiltersOperations, FiltersMethod, FiltersOrderByDirection,
    FiltersKnownFields as KnownFields,
    # triggers
    TriggerResource, TriggerAction, TriggerExecutionMode, TriggerType,
    # faas
    FunctionIO, KubernetesAutoscalerType, KubernetesRabbitmqAutoscaler, KubernetesAutoscaler, KubernetesRuntime,
    InstanceCatalog, PackageInputType, ServiceType, ServiceModeType, KubernetesRPSAutoscaler,
    PackageSlot, SlotPostAction, SlotPostActionType, SlotDisplayScope, SlotDisplayScopeResource, UiBindingPanel,
    # roberto
    DatasetSubsetType, ModelStatus, PlotSample, ArtifactType, Artifact, ItemArtifact, LinkArtifact, LocalArtifact,
    EntityScopeLevel,
    # features
    FeatureEntityType, Feature, FeatureSet,
    #
    RequirementOperator, PackageRequirement,
    Command, CommandsStatus,
    LocalCodebase, GitCodebase, ItemCodebase, FilesystemCodebase, PackageCodebaseType,
    MemberRole, MemberOrgRole,
    Webhook, HttpMethod,
    ViewAnnotationOptions, AnnotationStatus, AnnotationType,
    ItemStatus, ExecutionStatus, ExportMetadata,
    PromptItem, Prompt, PromptType, ItemGis, Layer,
    ItemLink, UrlLink, LinkTypeEnum,
    Modality, ModalityTypeEnum, ModalityRefTypeEnum,
    Workload, WorkloadUnit, ItemAction,
    PipelineExecution, PipelineExecutionStatus, CycleRerunMethod, PipelineExecutionNode, Pipeline, PipelineConnection,
    PipelineNode, TaskNode, CodeNode, PipelineStats, PipelineSettings,
    PipelineNodeType, PipelineNameSpace, PipelineResumeOption, Variable, CompositionStatus,
    FunctionNode, DatasetNode, PipelineConnectionPort, PipelineNodeIO, Organization, OrganizationsPlans, Integration,
    Driver, S3Driver, GcsDriver, AzureBlobDriver, CacheAction, PodType,
    ExternalStorage, IntegrationType, Role, PlatformEntityType, SettingsValueTypes, SettingsTypes, SettingsSectionNames,
    SettingScope, BaseSetting, UserSetting, Setting, ServiceSample, ExecutionSample, PipelineExecutionSample,
    ResourceExecution, Message, NotificationEventContext,
    # compute
    ClusterProvider, ComputeType, ComputeStatus, Toleration, DeploymentResource, DeploymentResources,
    NodePool, AuthenticationIntegration, Authentication, ComputeCluster, ComputeContext, Compute, KubernetesCompute,
    ServiceDriver, ExportType, OutputExportType, DynamicConcurrencyUpdateMethod
)
from .ml import BaseModelAdapter
from .utilities import Converter, BaseServiceRunner, Progress, Context, AnnotationFormat
from .repositories import FUNCTION_END_LINE, PackageCatalog

warnings.simplefilter('once', DeprecationWarning)

# check python version
if sys.version_info.major != 3:
    if sys.version_info.minor not in [5, 6]:
        sys.stderr.write(
            'Error: Your Python version "{}.{}" is NOT supported by Dataloop SDK dtlpy. '
            "Supported version are 3.5, 3.6)\n".format(
                sys.version_info.major, sys.version_info.minor
            )
        )
        sys.exit(-1)

if os.name == "nt":
    # set encoding for windows printing
    os.environ["PYTHONIOENCODING"] = ":replace"

"""
Main Platform Interface module for Dataloop
"""

################
# Repositories #
################
# Create repositories instances
projects = repositories.Projects(client_api=client_api)
datasets = repositories.Datasets(client_api=client_api, project=None)
items = repositories.Items(client_api=client_api, datasets=datasets)
packages = repositories.Packages(client_api=client_api)
executions = repositories.Executions(client_api=client_api)
commands = repositories.Commands(client_api=client_api)
services = repositories.Services(client_api=client_api)
webhooks = repositories.Webhooks(client_api=client_api)
triggers = repositories.Triggers(client_api=client_api)
assignments = repositories.Assignments(client_api=client_api)
tasks = repositories.Tasks(client_api=client_api)
annotations = repositories.Annotations(client_api=client_api)
models = repositories.Models(client_api=client_api)
ontologies = repositories.Ontologies(client_api=client_api)
recipes = repositories.Recipes(client_api=client_api)
pipelines = repositories.Pipelines(client_api=client_api)
pipeline_executions = repositories.PipelineExecutions(client_api=client_api)
feature_sets = repositories.FeatureSets(client_api=client_api)
features = repositories.Features(client_api=client_api)
organizations = repositories.Organizations(client_api=client_api)
analytics = repositories.Analytics(client_api=client_api)
integrations = repositories.Integrations(client_api=client_api)
drivers = repositories.Drivers(client_api=client_api)
settings = repositories.Settings(client_api=client_api)
apps = repositories.Apps(client_api=client_api)
dpks = repositories.Dpks(client_api=client_api)
messages = repositories.Messages(client_api=client_api)
compositions = repositories.Compositions(client_api=client_api)
computes = repositories.Computes(client_api=client_api)
service_drivers = repositories.ServiceDrivers(client_api=client_api)
collections = repositories.Collections(client_api=client_api)

try:
    check_sdk.check(version=__version__, client_api=client_api)
except Exception:
    logger.debug("Failed to check SDK! Continue without")

verbose = client_api.verbose
login = client_api.login
logout = client_api.logout
login_token = client_api.login_token
login_secret = client_api.login_secret
login_api_key = client_api.login_api_key
login_m2m = client_api.login_m2m
add_environment = client_api.add_environment
setenv = client_api.setenv
token_expired = client_api.token_expired
info = client_api.info
cache_state = client_api.cache_state
attributes_mode = client_api.attributes_mode
sdk_cache = client_api.sdk_cache


def get_secret(secret):
    return os.environ.get(secret, None)


def token():
    """
    token
    :return: token in use
    """
    return client_api.token


def environment():
    """
    environment
    :return: current environment
    """
    return client_api.environment


def init():
    """
    init current directory as a Dataloop working directory
    :return:
    """
    from .services import CookieIO

    client_api.state_io = CookieIO.init_local_cookie(create=True)
    assert isinstance(client_api.state_io, CookieIO)
    logger.info(".Dataloop directory initiated successfully in {}".format(os.getcwd()))


def checkout_state():
    """
    Return the current checked out state
    :return:
    """
    state = client_api.state_io.read_json()
    return state

class LoggingLevel:
    DEBUG = "debug"
    WARNING = "warning"
    CRITICAL = "critical"
    INFO = "info"
    ERROR = "error"


#################
#     ENUMS     #
#################
CallbackEvent = client_api.callbacks.CallbackEvent
LOGGING_LEVEL_DEBUG = LoggingLevel.DEBUG
LOGGING_LEVEL_WARNING = LoggingLevel.WARNING
LOGGING_LEVEL_CRITICAL = LoggingLevel.CRITICAL
LOGGING_LEVEL_INFO = LoggingLevel.INFO

VERBOSE_LOGGING_LEVEL_DEBUG = VerboseLoggingLevel.DEBUG
VERBOSE_LOGGING_LEVEL_INFO = VerboseLoggingLevel.INFO
VERBOSE_LOGGING_LEVEL_WARNING = VerboseLoggingLevel.WARNING
VERBOSE_LOGGING_LEVEL_ERROR = VerboseLoggingLevel.ERROR
VERBOSE_LOGGING_LEVEL_CRITICAL = VerboseLoggingLevel.CRITICAL

HTTP_METHOD_POST = HttpMethod.POST
HTTP_METHOD_GET = HttpMethod.GET
HTTP_METHOD_DELETE = HttpMethod.DELETE
HTTP_METHOD_PATCH = HttpMethod.PATCH

VIEW_ANNOTATION_OPTIONS_JSON = ViewAnnotationOptions.JSON
VIEW_ANNOTATION_OPTIONS_MASK = ViewAnnotationOptions.MASK
VIEW_ANNOTATION_OPTIONS_ANNOTATION_ON_IMAGE = ViewAnnotationOptions.ANNOTATION_ON_IMAGE
VIEW_ANNOTATION_OPTIONS_INSTANCE = ViewAnnotationOptions.INSTANCE
VIEW_ANNOTATION_OPTIONS_VTT = ViewAnnotationOptions.VTT
VIEW_ANNOTATION_OPTIONS_OBJECT_ID = ViewAnnotationOptions.OBJECT_ID

ANNOTATION_STATUS_ISSUE = AnnotationStatus.ISSUE
ANNOTATION_STATUS_REVIEW = AnnotationStatus.REVIEW
ANNOTATION_STATUS_APPROVED = AnnotationStatus.APPROVED
ANNOTATION_STATUS_CLEAR = AnnotationStatus.CLEAR

ORGANIZATION_PLAN_FREEMIUM = OrganizationsPlans.FREEMIUM
ORGANIZATION_PLAN_PREMIUM = OrganizationsPlans.PREMIUM

EXPORT_METADATA_FROM_JSON = ExportMetadata.FROM_JSON

EXPORT_VERSION_V1 = ExportVersion.V1
EXPORT_VERSION_V2 = ExportVersion.V2

# class
ANNOTATION_TYPE_BOX = AnnotationType.BOX
ANNOTATION_TYPE_CLASSIFICATION = AnnotationType.CLASSIFICATION
ANNOTATION_TYPE_COMPARISON = AnnotationType.COMPARISON
ANNOTATION_TYPE_ELLIPSE = AnnotationType.ELLIPSE
ANNOTATION_TYPE_NOTE = AnnotationType.NOTE
ANNOTATION_TYPE_POINT = AnnotationType.POINT
ANNOTATION_TYPE_POLYGON = AnnotationType.POLYGON
ANNOTATION_TYPE_POLYLINE = AnnotationType.POLYLINE
ANNOTATION_TYPE_POSE = AnnotationType.POSE
ANNOTATION_TYPE_SEGMENTATION = AnnotationType.SEGMENTATION
ANNOTATION_TYPE_SUBTITLE = AnnotationType.SUBTITLE
ANNOTATION_TYPE_TEXT = AnnotationType.TEXT
ANNOTATION_TYPE_FREE_TEXT = AnnotationType.FREE_TEXT
ANNOTATION_TYPE_REF_IMAGE = AnnotationType.REF_IMAGE

ITEM_STATUS_COMPLETED = ItemStatus.COMPLETED
ITEM_STATUS_APPROVED = ItemStatus.APPROVED
ITEM_STATUS_DISCARDED = ItemStatus.DISCARDED

EXECUTION_STATUS_CREATED = ExecutionStatus.CREATED
EXECUTION_STATUS_IN_PROGRESS = ExecutionStatus.IN_PROGRESS
EXECUTION_STATUS_SUCCESS = ExecutionStatus.SUCCESS
EXECUTION_STATUS_FAILED = ExecutionStatus.FAILED

LINK_TYPE_ID = LinkTypeEnum.ID
LINK_TYPE_URL = LinkTypeEnum.URL

KUBERNETES_AUTUSCALER_TYPE_CPU = KubernetesAutoscalerType.CPU
KUBERNETES_AUTUSCALER_TYPE_RABBITMQ = KubernetesAutoscalerType.RABBITMQ

INSTANCE_CATALOG_REGULAR_XS = InstanceCatalog.REGULAR_XS
INSTANCE_CATALOG_REGULAR_S = InstanceCatalog.REGULAR_S
INSTANCE_CATALOG_REGULAR_M = InstanceCatalog.REGULAR_M
INSTANCE_CATALOG_REGULAR_L = InstanceCatalog.REGULAR_L
INSTANCE_CATALOG_HIGHMEM_XS = InstanceCatalog.HIGHMEM_XS
INSTANCE_CATALOG_HIGHMEM_S = InstanceCatalog.HIGHMEM_S
INSTANCE_CATALOG_HIGHMEM_M = InstanceCatalog.HIGHMEM_M
INSTANCE_CATALOG_HIGHMEM_L = InstanceCatalog.HIGHMEM_L
INSTANCE_CATALOG_GPU_T4_S = InstanceCatalog.GPU_T4_S
INSTANCE_CATALOG_GPU_T4_M = InstanceCatalog.GPU_T4_M

MODALITY_TYPE_OVERLAY = ModalityTypeEnum.OVERLAY
MODALITY_TYPE_PREVIEW = ModalityTypeEnum.PREVIEW
MODALITY_TYPE_REPLACE = ModalityTypeEnum.REPLACE

MODALITY_REF_TYPE_ID = ModalityRefTypeEnum.ID
MODALITY_REF_TYPE_URL = ModalityRefTypeEnum.URL

FILTERS_KNOWN_FIELDS_DIR = FiltersKnownFields.DIR
FILTERS_KNOWN_FIELDS_ANNOTATED = FiltersKnownFields.ANNOTATED
FILTERS_KNOWN_FIELDS_FILENAME = FiltersKnownFields.FILENAME
FILTERS_KNOWN_FIELDS_CREATED_AT = FiltersKnownFields.CREATED_AT
FILTERS_KNOWN_FIELDS_UPDATED_AT = FiltersKnownFields.UPDATED_AT
FILTERS_KNOWN_FIELDS_LABEL = FiltersKnownFields.LABEL
FILTERS_KNOWN_FIELDS_NAME = FiltersKnownFields.NAME
FILTERS_KNOWN_FIELDS_HIDDEN = FiltersKnownFields.HIDDEN
FILTERS_KNOWN_FIELDS_TYPE = FiltersKnownFields.TYPE

FILTERS_RESOURCE_ITEM = FiltersResource.ITEM
FILTERS_RESOURCE_ANNOTATION = FiltersResource.ANNOTATION
FILTERS_RESOURCE_EXECUTION = FiltersResource.EXECUTION
FILTERS_RESOURCE_PACKAGE = FiltersResource.PACKAGE
FILTERS_RESOURCE_SERVICE = FiltersResource.SERVICE
FILTERS_RESOURCE_TRIGGER = FiltersResource.TRIGGER
FILTERS_RESOURCE_MODEL = FiltersResource.MODEL
FILTERS_RESOURCE_WEBHOOK = FiltersResource.WEBHOOK
FILTERS_RESOURCE_RECIPE = FiltersResource.RECIPE
FILTERS_RESOURCE_DATASET = FiltersResource.DATASET
FILTERS_RESOURCE_ONTOLOGY = FiltersResource.ONTOLOGY

FILTERS_OPERATIONS_OR = FiltersOperations.OR
FILTERS_OPERATIONS_AND = FiltersOperations.AND
FILTERS_OPERATIONS_IN = FiltersOperations.IN
FILTERS_OPERATIONS_NOT_EQUAL = FiltersOperations.NOT_EQUAL
FILTERS_OPERATIONS_EQUAL = FiltersOperations.EQUAL
FILTERS_OPERATIONS_GREATER_THAN = FiltersOperations.GREATER_THAN
FILTERS_OPERATIONS_LESS_THAN = FiltersOperations.LESS_THAN
FILTERS_OPERATIONS_EXISTS = FiltersOperations.EXISTS

FILTERS_METHOD_OR = FiltersMethod.OR
FILTERS_METHOD_AND = FiltersMethod.AND

FILTERS_ORDERBY_DIRECTION_DESCENDING = FiltersOrderByDirection.DESCENDING
FILTERS_ORDERBY_DIRECTION_ASCENDING = FiltersOrderByDirection.ASCENDING

TRIGGER_RESOURCE_ITEM = TriggerResource.ITEM
TRIGGER_RESOURCE_DATASET = TriggerResource.DATASET
TRIGGER_RESOURCE_ANNOTATION = TriggerResource.ANNOTATION
TRIGGER_RESOURCE_ITEM_STATUS = TriggerResource.ITEM_STATUS

TRIGGER_ACTION_CREATED = TriggerAction.CREATED
TRIGGER_ACTION_UPDATED = TriggerAction.UPDATED
TRIGGER_ACTION_DELETED = TriggerAction.DELETED
TRIGGER_ACTION_STATUS_CHANGED = TriggerAction.STATUS_CHANGED
TRIGGER_ACTION_CLONE = TriggerAction.CLONE

TRIGGER_EXECUTION_MODE_ONCE = TriggerExecutionMode.ONCE
TRIGGER_EXECUTION_MODE_ALWAYS = TriggerExecutionMode.ALWAYS

TRIGGER_TYPE_EVENT = TriggerType.EVENT
TRIGGER_TYPE_CRON = TriggerType.CRON

PACKAGE_INPUT_TYPE_ITEM = PackageInputType.ITEM
PACKAGE_INPUT_TYPE_DATASET = PackageInputType.DATASET
PACKAGE_INPUT_TYPE_ANNOTATION = PackageInputType.ANNOTATION
PACKAGE_INPUT_TYPE_JSON = PackageInputType.JSON
PACKAGE_INPUT_TYPE_MODEL = PackageInputType.MODEL
PACKAGE_INPUT_TYPE_PACKAGE = PackageInputType.PACKAGE
PACKAGE_INPUT_TYPE_SERVICE = PackageInputType.SERVICE
PACKAGE_INPUT_TYPE_PROJECT = PackageInputType.PROJECT
PACKAGE_INPUT_TYPE_EXECUTION = PackageInputType.EXECUTION
PACKAGE_INPUT_TYPE_TASK = PackageInputType.TASK
PACKAGE_INPUT_TYPE_ASSIGNMENT = PackageInputType.ASSIGNMENT
PACKAGE_INPUT_TYPE_RECIPE = PackageInputType.RECIPE
PACKAGE_INPUT_TYPE_STRING = PackageInputType.STRING
PACKAGE_INPUT_TYPE_NUMBER = PackageInputType.NUMBER
PACKAGE_INPUT_TYPE_INT = PackageInputType.INT
PACKAGE_INPUT_TYPE_FLOAT = PackageInputType.FLOAT
PACKAGE_INPUT_TYPE_BOOLEAN = PackageInputType.BOOLEAN
PACKAGE_INPUT_TYPE_DATASETS = PackageInputType.DATASETS
PACKAGE_INPUT_TYPE_ITEMS = PackageInputType.ITEMS
PACKAGE_INPUT_TYPE_ANNOTATIONS = PackageInputType.ANNOTATIONS
PACKAGE_INPUT_TYPE_EXECUTIONS = PackageInputType.EXECUTIONS
PACKAGE_INPUT_TYPE_TASKS = PackageInputType.TASKS
PACKAGE_INPUT_TYPE_ASSIGNMENTS = PackageInputType.ASSIGNMENTS
PACKAGE_INPUT_TYPE_SERVICES = PackageInputType.SERVICES
PACKAGE_INPUT_TYPE_PACKAGES = PackageInputType.PACKAGES
PACKAGE_INPUT_TYPE_PROJECTS = PackageInputType.PROJECTS
PACKAGE_INPUT_TYPE_JSONS = PackageInputType.JSONS
PACKAGE_INPUT_TYPE_STRINGS = PackageInputType.STRINGS
PACKAGE_INPUT_TYPE_NUMBERS = PackageInputType.NUMBERS
PACKAGE_INPUT_TYPE_INTS = PackageInputType.INTS
PACKAGE_INPUT_TYPE_FLOATS = PackageInputType.FLOATS
PACKAGE_INPUT_TYPE_BOOLEANS = PackageInputType.BOOLEANS
PACKAGE_INPUT_TYPE_MODELS = PackageInputType.MODELS
PACKAGE_INPUT_TYPE_RECIPES = PackageInputType.RECIPES

FUNCTION_POST_ACTION_TYPE_DOWNLOAD = SlotPostActionType.DOWNLOAD
FUNCTION_POST_ACTION_TYPE_DRAW_ANNOTATION = SlotPostActionType.DRAW_ANNOTATION
FUNCTION_POST_ACTION_TYPE_NO_ACTION = SlotPostActionType.NO_ACTION

FUNCTION_DISPLAY_SCOPE_RESOURCE_ANNOTATION = SlotDisplayScopeResource.ANNOTATION
FUNCTION_DISPLAY_SCOPE_RESOURCE_ITEM = SlotDisplayScopeResource.ITEM
FUNCTION_DISPLAY_SCOPE_RESOURCE_DATASET = SlotDisplayScopeResource.DATASET
FUNCTION_DISPLAY_SCOPE_RESOURCE_DATASET_QUERY = SlotDisplayScopeResource.DATASET_QUERY
FUNCTION_DISPLAY_SCOPE_RESOURCE_TASK = SlotDisplayScopeResource.TASK

UI_BINDING_PANEL_ALL = UiBindingPanel.ALL
UI_BINDING_PANEL_TABLE = UiBindingPanel.TABLE
UI_BINDING_PANEL_STUDIO = UiBindingPanel.STUDIO
UI_BINDING_PANEL_BROWSER = UiBindingPanel.BROWSER

COMMANDS_STATUS_CREATED = CommandsStatus.CREATED
COMMANDS_STATUS_MAKING_CHILDREN = CommandsStatus.MAKING_CHILDREN
COMMANDS_STATUS_WAITING_CHILDREN = CommandsStatus.WAITING_CHILDREN
COMMANDS_STATUS_IN_PROGRESS = CommandsStatus.IN_PROGRESS
COMMANDS_STATUS_ABORTED = CommandsStatus.ABORTED
COMMANDS_STATUS_CANCELED = CommandsStatus.CANCELED
COMMANDS_STATUS_FINALIZING = CommandsStatus.FINALIZING
COMMANDS_STATUS_SUCCESS = CommandsStatus.SUCCESS
COMMANDS_STATUS_FAILED = CommandsStatus.FAILED
COMMANDS_STATUS_TIMEOUT = CommandsStatus.TIMEOUT

MEMBER_ROLE_OWNER = MemberRole.OWNER
MEMBER_ROLE_DEVELOPER = MemberRole.DEVELOPER
MEMBER_ROLE_ANNOTATOR = MemberRole.ANNOTATOR
MEMBER_ROLE_ANNOTATION_MANAGER = MemberRole.ANNOTATION_MANAGER

MEMBER_ORG_ROLE_OWNER = MemberOrgRole.OWNER
MEMBER_ORG_ROLE_ADMIN = MemberOrgRole.ADMIN
MEMBER_ORG_ROLE_MEMBER = MemberOrgRole.MEMBER
MEMBER_ORG_ROLE_WORKER = MemberOrgRole.WORKER

PACKAGE_REQUIREMENT_OP_EQUAL = RequirementOperator.EQUAL
PACKAGE_REQUIREMENT_OP_GREATER_THAN = RequirementOperator.GREATER_THAN
PACKAGE_REQUIREMENT_OP_LESS_THAN = RequirementOperator.LESS_THAN
PACKAGE_REQUIREMENT_OP_EQUAL_OR_LESS_THAN = RequirementOperator.EQUAL_OR_LESS_THAN
PACKAGE_REQUIREMENT_OP_EQUAL_OR_GREATER_THAN = RequirementOperator.EQUAL_OR_GREATER_THAN

INDEX_DRIVER_V1 = IndexDriver.V1
INDEX_DRIVER_V2 = IndexDriver.V2

CACHE_ACTION_APPLY = CacheAction.APPLY
CACHE_ACTION_DESTROY = CacheAction.DESTROY

POD_TYPE_SMALL = PodType.SMALL
POD_TYPE_MEDIUM = PodType.MEDIUM
POD_TYPE_HIGH = PodType.HIGH

ATTRIBUTES_TYPES_CHECKBOX = AttributesTypes.CHECKBOX
ATTRIBUTES_TYPES_RADIO_BUTTON = AttributesTypes.RADIO_BUTTON
ATTRIBUTES_TYPES_YES_NO = AttributesTypes.YES_NO
ATTRIBUTES_TYPES_SLIDER = AttributesTypes.SLIDER
ATTRIBUTES_TYPES_FREE_TEXT = AttributesTypes.FREE_TEXT

TASK_PRIORITY_LOW = TaskPriority.LOW
TASK_PRIORITY_MEDIUM = TaskPriority.MEDIUM
TASK_PRIORITY_HIGH = TaskPriority.HIGH

CONSENSUS_TASK_TYPE_CONSENSUS = ConsensusTaskType.CONSENSUS
CONSENSUS_TASK_TYPE_QUALIFICATION = ConsensusTaskType.QUALIFICATION
CONSENSUS_TASK_TYPE_HONEYPOT = ConsensusTaskType.HONEYPOT

TASK_TYPE_PULLING = AllocationMethod.PULLING
TASK_TYPE_DISTRIBUTION = AllocationMethod.DISTRIBUTION

SERVICE_MODE_TYPE_REGULAR = ServiceModeType.REGULAR
SERVICE_MODE_TYPE_DEBUG = ServiceModeType.DEBUG



import json
import warnings
from enum import Enum
from .. import repositories


class Role(str, Enum):
    OWNER = "owner",
    ADMIN = "admin",
    MEMBER = "member",
    ANNOTATOR = "annotator",
    DEVELOPER = "engineer",
    ANNOTATION_MANAGER = "annotationManager"
    ALL = "*"


class PlatformEntityType(str, Enum):
    USER = "user",
    TASK = "task",
    PROJECT = "project",
    ORG = "org",
    DATASET = "dataset"
    DATALOOP = "DATALOOP"


class SettingsValueTypes(str, Enum):
    BOOLEAN = "boolean",
    NUMBER = "number",
    SELECT = "select",
    MULTI_SELECT = "multi-select"
    STRING = "string"


class SettingsTypes(str, Enum):
    FEATURE_FLAG = "feature_flag",
    USER_SETTINGS = "user_settings"


class SettingsSectionNames(str, Enum):
    ACCOUNT = "Account",
    CONTACT = "Contact",
    APPLICATIONS = "Applications",
    STUDIO = "Studio",
    PLATFORM = "Platform"
    SDK = "SDK"


class SettingScope:
    def __init__(
            self,
            type: PlatformEntityType,
            id: str,
            role: Role,
            prevent_override: bool,
            visible: bool,
    ):
        self.type = type
        self.id = id
        self.role = role
        self.prevent_override = prevent_override
        self.visible = visible

    @staticmethod
    def from_json(_json):
        return SettingScope(
            type=_json.get('type', None),
            id=_json.get('id', None),
            role=_json.get('role', None),
            prevent_override=_json.get('preventOverride', None),
            visible=_json.get('visible', None)
        )

    def to_json(self):
        _json = dict()
        _json['type'] = self.type
        _json['id'] = self.id
        _json['role'] = self.role
        _json['preventOverride'] = self.prevent_override
        _json['visible'] = self.visible

        return _json


class BaseSetting:

    def __init__(
            self,
            default_value,
            value,
            name: str,
            value_type: SettingsValueTypes,
            scope: SettingScope,
            metadata: dict,
            setting_type: SettingsTypes,
            id: str = None,
            client_api=None,
            project=None,
            org=None
    ):

        self.default_value = default_value
        self.name = name
        self.value = value
        self.value_type = value_type
        self.scope = scope
        self.metadata = metadata
        self.setting_type = setting_type
        self.client_api = client_api
        self.project = project
        self.org = org
        self.id = id
        self.settings = repositories.Settings(client_api=self.client_api, project=self.project, org=self.org)

    @staticmethod
    def from_json(_json: dict, client_api, project=None, org=None):
        scope = SettingScope.from_json(_json.get('scope', None))
        return BaseSetting(
            default_value=_json.get('defaultValue', None),
            name=_json.get('name', None),
            value=_json.get('value', None),
            value_type=_json.get('valueType', None),
            scope=scope,
            metadata=_json.get('metadata', None),
            id=_json.get('id', None),
            setting_type=_json.get('settingType', None),
            client_api=client_api,
            project=project,
            org=org
        )

    @staticmethod
    def __slot_to_db_slot(slot: dict):
        if 'displayScopes' in slot:
            scopes = slot['displayScopes']
            if scopes:
                for display_scope in scopes:
                    if 'filter' in display_scope and isinstance(display_scope['filter'], dict):
                        display_scope['filter'] = json.dumps(display_scope['filter'])

    def to_json(self):
        if self.metadata is not None and 'slots' in self.metadata and isinstance(self.metadata['slots'], list):
            for slot in self.metadata['slots']:
                self.__slot_to_db_slot(slot)

        _json = {
            'name': self.name,
            'valueType': self.value_type,
            'scope': self.scope.to_json(),
            'settingType': self.setting_type,
            'id': self.id,
            'metadata': self.metadata,
            'value': self.value,
            'defaultValue': self.default_value
        }

        return _json

    def delete(self):
        """
        Delete a setting

        :return: True if success exceptions if not
        """
        return self.settings.delete(setting_id=self.id)

    def update(self):
        """
        Update a setting

        :return: setting entity
        """
        return self.settings.update(setting=self)


class Setting(BaseSetting):
    def __init__(
            self,
            value,
            name: str,
            value_type: SettingsValueTypes,
            scope: SettingScope,
            section_name: SettingsSectionNames,
            default_value=None,
            inputs=None,
            metadata: dict = None,
            description: str = None,
            icon: str = None,
            id: str = None,
            sub_section_name: str = None,
            hint=None,
            client_api=None,
            project=None,
            org=None,
            setting_type=SettingsTypes.USER_SETTINGS
    ):
        super().__init__(
            default_value=default_value,
            value=value,
            name=name,
            value_type=value_type,
            scope=scope,
            metadata=metadata,
            setting_type=setting_type,
            client_api=client_api,
            project=project,
            org=org,
            id=id
        )
        self.description = description
        self.inputs = inputs
        self.icon = icon
        self.section_name = section_name
        self.hint = hint
        self.sub_section_name = sub_section_name

    @staticmethod
    def from_json(_json: dict, client_api, project=None, org=None):
        scope = SettingScope.from_json(_json.get('scope', None))
        setting = Setting(
            default_value=_json.get('defaultValue', None),
            name=_json.get('name', None),
            value=_json.get('value', None),
            value_type=_json.get('valueType', None),
            scope=scope,
            metadata=_json.get('metadata', None),
            description=_json.get('description', None),
            inputs=_json.get('inputs', None),
            icon=_json.get('icon', None),
            section_name=_json.get('sectionName', None),
            hint=_json.get('hint', None),
            sub_section_name=_json.get('subSectionName', None),
            id=_json.get('id', None),
            client_api=client_api,
            project=project,
            org=org
        )
        if setting.metadata is not None and 'slots' in setting.metadata:
            for slot in setting.metadata.get('slots', []):
                for scope in slot.get('displayScopes', []):
                    if 'filter' in scope:
                        scope['filter'] = json.loads(scope['filter']) if isinstance(scope['filter'], str) else scope['filter']

        return setting

    def to_json(self):
        _json = super().to_json()
        _json['description'] = self.description
        _json['inputs'] = self.inputs
        _json['icon'] = self.icon
        _json['sectionName'] = self.section_name
        _json['hint'] = self.hint
        _json['subSectionName'] = self.sub_section_name
        _json['id'] = self.id
        return _json


class UserSetting(Setting):
    def __init__(
            self,
            default_value,
            value,
            name: str,
            value_type: SettingsValueTypes,
            scope: SettingScope,
            section_name: SettingsSectionNames,
            inputs=None,
            metadata: dict = None,
            description: str = None,
            icon: str = None,
            id: str = None,
            sub_section_name: str = None,
            hint=None,
            client_api=None,
            project=None,
            org=None
    ):
        super().__init__(
            default_value=default_value,
            value=value,
            name=name,
            value_type=value_type,
            scope=scope,
            metadata=metadata,
            client_api=client_api,
            project=project,
            org=org,
            id=id,
            section_name=section_name,
            inputs=inputs,
            description=description,
            icon=icon,
            sub_section_name=sub_section_name,
            hint=hint
        )



import dtlpy as dl
import logging

logger = logging.getLogger(name='dtlpy')


class ServiceRunner(dl.BaseServiceRunner):
    """
    Package runner class

    """

    def __init__(self):
        """
        Init package attributes here

        :return:
        """

    def run(self, config, progress=None):
        """
        In this example we copy dataset items by query to a new dataset

        :param config: This is a json input
        :param progress: Use this to update the progress of your package
        :return:
        """
        # these lines can be removed
        assert isinstance(progress, dl.Progress)
        source_dataset = dl.datasets.get(dataset_id=config['source_dataset_id'])
        filters = dl.Filters(custom_filter=config['query'])
        new_dataset = source_dataset.project.datasets.create(dataset_name='{}-copy'.format(source_dataset.name),
                                                             labels=source_dataset.labels)
        new_dataset.items.upload(local_path=source_dataset.items.download(filters=filters))
        logger.info('Dataset copied successfully')


if __name__ == "__main__":
    """
    Run this main to locally debug your package
    """
    dl.packages.test_local_package()



import logging
import pandas as pd
import datetime
from dtlpy import entities, miscellaneous, exceptions
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class TimesSeries:
    """
    Time series Repository
    """

    def __init__(self, client_api: ApiClient, project: entities.Project = None):
        self._client_api = client_api
        self._project = project

    ############
    # entities #
    ############
    @property
    def project(self) -> entities.Project:
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "project". need to set a Project entity or use project.times_series repository')
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    ############
    #  methods #
    ############
    def create(self, series_name) -> entities.TimeSeries:
        """
        Create a new time series

        :param str series_name: name
        :return: TimeSeries object
        """
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/projects/{}/timeSeries'.format(self.project.id),
                                                         json_req={'name': series_name})
        if success:
            ts = entities.TimeSeries.from_json(_json=response.json(),
                                               project=self.project)
        else:
            raise exceptions.PlatformException(response)
        assert isinstance(ts, entities.TimeSeries)
        return ts

    def list(self) -> miscellaneous.List[entities.TimeSeries]:
        """
        List all time series for project

        :return:
        """
        success, response = self._client_api.gen_request(req_type='get',
                                                         path='/projects/{}/timeSeries'.format(self.project.id))
        if success:
            tss = miscellaneous.List([entities.TimeSeries.from_json(_json=_json, project=self.project)
                                      for _json in response.json()])
        else:
            raise exceptions.PlatformException(response)
        return tss

    def get(self, series_name=None, series_id=None) -> entities.TimeSeries:
        """
        Get time series entity

        :param str series_name: by name
        :param str series_id: by id
        :return:
        """
        if series_id is not None:
            # get series
            success, response = self._client_api.gen_request(req_type='get',
                                                             path='/projects/{}/timeSeries/{}'.format(self.project.id,
                                                                                                      series_id))
            if success:
                ts = entities.TimeSeries.from_json(_json=response.json(),
                                                   project=self.project)
            else:
                raise exceptions.PlatformException(response)
            # verify input service name is same as the given id
            if series_name is not None and ts.name != series_name:
                logger.warning(
                    "Mismatch found in timeSeries.get: series_name is different then timeSeries.name:"
                    " {!r} != {!r}".format(
                        series_name,
                        ts.name))
        elif series_name is not None:
            tss = self.list()
            ts = [ts for ts in tss if ts.name == series_name]
            if not ts:
                # empty list
                raise exceptions.PlatformException(error='404',
                                                   message='Time Series not found. Name: {}'.format(series_name))
            elif len(ts) > 1:
                raise exceptions.PlatformException(error='400',
                                                   message='More than one Time Series with same name.')
            else:
                ts = ts[0]
        else:
            raise exceptions.PlatformException(error='400',
                                               message='Must choose by "series_id" or "series_name"')
        assert isinstance(ts, entities.TimeSeries)
        return ts

    def delete(self, series_id=None, series=None):
        """
        Delete a Time Series

        :param str series_id: optional - search by id
        :param series: optional - TimeSeries object
        :return: True
        :rtype: bool
        """
        if series_id is not None:
            pass
        elif series is not None and isinstance(series, entities.TimeSeries):
            series_id = series.id
        else:
            msg = 'Must choose by at least one of: "series_id", "series"'
            logger.error(msg)
            raise ValueError(msg)
        success, response = self._client_api.gen_request(req_type='delete',
                                                         path='/projects/{}/timeSeries/{}'.format(self.project.id,
                                                                                                  series_id))
        if not success:
            raise exceptions.PlatformException(response)
        logger.info('Time series id {} deleted successfully'.format(series_id))
        return True

    #########
    # Table #
    #########
    def delete_samples(self, series_id, filters):
        """
        Delete samples from table

        :param str series_id: time series id
        :param dtlpy.entities.filters.Filters filters: query to delete by
        :return:
        """
        filters = self._validate_query(query=filters)
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/projects/{}/timeSeries/{}/remove'.format(
                                                             self.project.id,
                                                             series_id),
                                                         json_req=filters)
        if not success:
            raise exceptions.PlatformException(response)
        return True

    @staticmethod
    def _validate_query(query):
        default_start_time = 0
        default_end_time = datetime.datetime.now().timestamp() * 1000

        if query is None:
            query = dict(startTime=default_start_time, endTime=default_end_time)
        else:
            query['startTime'] = query.get('startTime', default_start_time)
            query['endTime'] = query.get('endTime', default_end_time)

        return query

    def get_samples(self, series_id, filters=None) -> pd.DataFrame:
        """
        Get Series table

        :param str series_id: TimeSeries id
        :param dtlpy.entities.filters.Filters filters: match filters to get specific data from series
        :return:
        """
        filters = self._validate_query(query=filters)
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/projects/{}/timeSeries/{}/query'.format(self.project.id,
                                                                                                        series_id),
                                                         json_req=filters)
        if success:
            res = response.json()['samples']
            if isinstance(res, dict):
                df = pd.DataFrame([res])
            elif isinstance(res, list):
                df = pd.DataFrame(res)
            else:
                raise ValueError('unknown return type for time series: {}'.format(type(res)))
        else:
            raise exceptions.PlatformException(response)
        return df

    def add_samples(self, series_id, data):
        """
        Add samples to series

        :param str series_id: TimeSeries id
        :param data: list or dictionary of samples
        :return:
        """
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/projects/{}/timeSeries/{}/add'.format(self.project.id,
                                                                                                      series_id),
                                                         json_req=data)

        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    ######################
    # Samples Operations #
    ######################
    def get_sample(self, series_id, sample_id) -> pd.DataFrame:
        """
        Get single sample from series

        :param str series_id: TimeSeries id
        :param str sample_id: id of sample line
        :return:
        """
        success, response = self._client_api.gen_request(req_type='get',
                                                         path='/projects/{}/timeSeries/{}/samples/{}'.format(
                                                             self.project.id,
                                                             series_id,
                                                             sample_id))
        if success:
            res = response.json()
            if isinstance(res, dict):
                df = pd.DataFrame([res])
            elif isinstance(res, list):
                df = pd.DataFrame(res)
            else:
                raise ValueError('unknown return type for time series: {}'.format(type(res)))
        else:
            raise exceptions.PlatformException(response)
        return df

    def update_sample(self, series_id, sample_id, data):
        """
        Add data to existing sample

        :param str series_id: time series id
        :param str sample_id: sample line id
        :param dict data: dictionary
        :return:
        """
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/projects/{}/timeSeries/{}/samples/{}'.format(
                                                             self.project.id,
                                                             series_id,
                                                             sample_id),
                                                         json_req=data)
        if not success:
            raise exceptions.PlatformException(response)

    def delete_sample(self, series_id, sample_id):
        """
        Delete single samples form time series

        :param str series_id:
        :param str sample_id:
        :return:
        """
        success, response = self._client_api.gen_request(req_type='delete',
                                                         path='/projects/{}/timeSeries/{}/samples/{}'.format(
                                                             self.project.id,
                                                             series_id,
                                                             sample_id))
        if not success:
            raise exceptions.PlatformException(response)
        return True



from enum import Enum
import logging
import attr

from .. import entities, exceptions, repositories
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class IntegrationType(str, Enum):
    """ The type of the Integration.

    .. list-table::
       :widths: 15 150
       :header-rows: 1

       * - State
         - Description
       * - S3
         - S3 Integration - for S3 drivers
       * - AWS_CROSS_ACCOUNT
         - AWS CROSS ACCOUNT Integration - for S3 drivers
       * - AWS_STS
         - AWS STS Integration - for S3 drivers
       * - GCS
         - GCS Integration - for GCS drivers
       * - GCP_CROSS_PROJECT
         - GCP CROSS PROJECT Integration - for GCP drivers
       * - AZUREBLOB
         - AZURE BLOB Integration - for S3 AZUREBLOB and AZURE_DATALAKE_GEN2 drivers
       * - KEY_VALUE
         - KEY VALUE Integration - for save secrets in the platform
       * - GCP_WORKLOAD_IDENTITY_FEDERATION
         - GCP Workload Identity Federation Integration - for GCP drivers
       * - PRIVATE_REGISTRY
         - PRIVATE REGISTRY Integration - for private registry drivers
    """
    S3 = "s3"
    AWS_CROSS_ACCOUNT = 'aws-cross'
    AWS_STS = 'aws-sts'
    GCS = "gcs"
    GCS_CROSS = "gcp-cross"
    AZUREBLOB = "azureblob"
    KEY_VALUE = "key_value"
    GCP_WORKLOAD_IDENTITY_FEDERATION = "gcp-workload-identity-federation",
    PRIVATE_REGISTRY = "private-registry"


@attr.s
class Integration(entities.BaseEntity):
    """
    Integration object
    """
    id = attr.ib()
    name = attr.ib()
    type = attr.ib()
    org = attr.ib()
    created_at = attr.ib()
    creator = attr.ib()
    update_at = attr.ib()
    url = attr.ib()
    _client_api = attr.ib(type=ApiClient, repr=False)
    metadata = attr.ib(default=None, repr=False)
    _project = attr.ib(default=None, repr=False)

    @classmethod
    def from_json(cls,
                  _json: dict,
                  client_api: ApiClient,
                  is_fetched=True):
        """
        Build a Integration entity object from a json

        :param _json: _json response from host
        :param client_api: ApiClient entity
        :param is_fetched: is Entity fetched from Platform
        :return: Integration object
        """
        inst = cls(id=_json.get('id', None),
                   name=_json.get('name', None),
                   creator=_json.get('creator', None),
                   created_at=_json.get('createdAt', None),
                   update_at=_json.get('updatedAt', None),
                   type=_json.get('type', None),
                   org=_json.get('org', None),
                   client_api=client_api,
                   metadata=_json.get('metadata', None),
                   url=_json.get('url', None))
        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        _json = attr.asdict(self, filter=attr.filters.exclude(attr.fields(Integration)._client_api,
                                                              attr.fields(Integration)._project))
        return _json

    @property
    def project(self):
        return self._project

    @project.setter
    def project(self, project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    def update(
            self,
            new_name: str = None,
            new_options: dict = None,
            reload_services: bool = None
    ):
        """
        Update the integration's name.

        **Prerequisites**: You must be an *owner* in the organization.

        :param str new_name: new name
        :param dict new_options: new value
        :param bool reload_services: reload services associated with this integration
        :return: Integration object
        :rtype: dtlpy.entities.integration.Integration

        **Examples for options include**:
        s3 - {key: "", secret: ""};
        gcs - {key: "", secret: "", content: ""};
        azureblob - {key: "", secret: "", clientId: "", tenantId: ""};
        key_value - {key: "", value: ""}
        aws-sts - {key: "", secret: "", roleArns: ""}
        aws-cross - {roleArns: ""}

        **Example**:

        .. code-block:: python

            project.integrations.update(integrations_id='integrations_id', new_name="new_integration_name")
        """
        if self.project is not None:
            identifier = self.project
        elif self.org is not None:
            identifier = repositories.organizations.Organizations(client_api=self._client_api).get(
                organization_id=self.org)
        else:
            raise exceptions.PlatformException(
                error='400',
                message='Must provide an identifier in inputs')

        identifier.integrations.update(
            new_name=new_name,
            integrations_id=self.id,
            integration=self,
            new_options=new_options,
            reload_services=reload_services
        )

    def delete(self,
               sure: bool = False,
               really: bool = False) -> bool:
        """
        Delete integrations from the Organization

        :param bool sure: are you sure you want to delete?
        :param bool really: really really?
        :return: True
        :rtype: bool
        """
        if self.project is not None:
            identifier = self.project
        elif self.org is not None:
            identifier = repositories.organizations.Organizations(client_api=self._client_api).get(
                organization_id=self.org)
        else:
            raise exceptions.PlatformException(
                error='400',
                message='Must provide an identifier in inputs')

        return identifier.integrations.delete(integrations_id=self.id, sure=sure, really=really)



def main():
    import dtlpy as dl
    import numpy as np
    import matplotlib.pyplot as plt

    # Get project and dataset
    project = dl.projects.get(project_name='Food')
    dataset = project.datasets.get(dataset_name='BeansDataset')

    # get item from platform
    item = dataset.items.get(filepath='/image.jpg')

    # Create a builder instance
    builder = item.annotations.builder()

    # add annotations of type box and label person
    builder.add(annotation_definition=dl.Box(top=10,
                                             left=10,
                                             bottom=100,
                                             right=100,
                                             label='black_bean'))

    # add annotations of type point with attribute
    builder.add(annotation_definition=dl.Point(x=80,
                                               y=80,
                                               label='pea'),
                attribute=['large'])

    # add annotations of type polygon
    builder.add(annotation_definition=dl.Polyline(geo=[[80, 40],
                                                       [100, 120],
                                                       [110, 130]],
                                                  label='beans_can'))

    # create a mask
    mask = np.zeros(shape=(item.height, item.width), dtype=np.uint8)
    # mark some part in the middle
    mask[50:100, 200:250] = 1
    # add annotations of type segmentation
    builder.add(annotation_definition=dl.Segmentation(geo=mask,
                                                      label='tomato_sauce'))

    # plot the all of the annotations you created
    plt.figure()
    plt.imshow(builder.show())
    for annotation in builder:
        plt.figure()
        plt.imshow(annotation.show())
        plt.title(annotation.label)
    # upload annotations to the item
    item.annotations.upload(builder)



import logging

from .. import entities, exceptions, repositories, miscellaneous, _api_reference
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Items:
    """
    Items Repository

    The Items class allows you to manage items in your datasets.
    For information on actions related to items see https://developers.dataloop.ai/tutorials/data_management/upload_and_manage_items/chapter/
    """

    def __init__(self,
                 client_api: ApiClient,
                 datasets: repositories.Datasets = None,
                 dataset: entities.Dataset = None,
                 dataset_id=None,
                 items_entity=None,
                 project=None):
        self._client_api = client_api
        self._dataset = dataset
        self._dataset_id = dataset_id
        self._datasets = datasets
        self._project = project
        # set items entity to represent the item (Item, Codebase, Artifact etc...)
        if items_entity is None:
            self.items_entity = entities.Item
        if self._dataset_id is None and self._dataset is not None:
            self._dataset_id = self._dataset.id

    ############
    # entities #
    ############
    @property
    def dataset(self) -> entities.Dataset:
        if self._dataset is None:
            if self._dataset_id is None:
                raise exceptions.PlatformException(
                    error='400',
                    message='Cannot perform action WITHOUT Dataset entity in Items repository. Please set a dataset')
            self._dataset = self.datasets.get(dataset_id=self._dataset_id, fetch=None)
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @dataset.setter
    def dataset(self, dataset: entities.Dataset):
        if not isinstance(dataset, entities.Dataset):
            raise ValueError('Must input a valid Dataset entity')
        self._dataset = dataset

    @property
    def project(self) -> entities.Project:
        if self._project is None:
            raise exceptions.PlatformException(
                error='400',
                message='Cannot perform action WITHOUT Project entity in Items repository. Please set a project')
        assert isinstance(self._dataset, entities.Dataset)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Dataset entity')
        self._project = project

    ################
    # repositories #
    ################
    @property
    def datasets(self) -> repositories.Datasets:
        if self._datasets is None:
            self._datasets = repositories.Datasets(client_api=self._client_api)
        assert isinstance(self._datasets, repositories.Datasets)
        return self._datasets

    ###########
    # methods #
    ###########

    def set_items_entity(self, entity):
        """
        Set the item entity type to `Artifact <https://dataloop.ai/docs/auto-annotation-service?#uploading-model-weights-as-artifacts>`_, Item, or Codebase.

        :param entities.Item, entities.Artifact, entities.Codebase entity: entity type [entities.Item, entities.Artifact, entities.Codebase]
        """
        if entity in [entities.Item, entities.Artifact, entities.Codebase]:
            self.items_entity = entity
        else:
            raise exceptions.PlatformException(error="403",
                                               message="Unable to set given entity. Entity give: {}".format(entity))

    def get_all_items(self, filters: entities.Filters = None) -> [entities.Item]:
        """
        Get all items in dataset.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.filters.Filters filters: dl.Filters entity to filters items
        :return: list of all items
        :rtype: list

        **Example**:

        .. code-block:: python

            dataset.items.get_all_items()

        """
        if filters is None:
            filters = entities.Filters()
            filters._user_query = 'false'
            filters.add(field='type', values='file')
        pages = self.list(filters=filters)
        num_items = pages.items_count
        items = [None for _ in range(num_items)]
        for i_item, item in enumerate(pages.all()):
            items[i_item] = item
        items = [item for item in items if item is not None]
        return items

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Item]:
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        # return triggers list
        for i_item, item in enumerate(response_items):
            jobs[i_item] = pool.submit(self.items_entity._protected_from_json,
                                       **{'client_api': self._client_api,
                                          '_json': item,
                                          'dataset': self.dataset})
        # get all results
        results = [j.result() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        items = miscellaneous.List([r[1] for r in results if r[0] is True])
        return items

    def _list(self, filters: entities.Filters):
        """
        Get dataset items list This is a browsing endpoint, for any given path item count will be returned,
        user is expected to perform another request then for every folder item to actually get the its item list.

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: json response
        """
        # prepare request
        success, response = self._client_api.gen_request(req_type="POST",
                                                         path="/datasets/{}/query".format(self.dataset.id),
                                                         json_req=filters.prepare(),
                                                         headers={'user_query': filters._user_query})
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    @_api_reference.add(path='/datasets/{id}/query', method='post')
    def list(self,
             filters: entities.Filters = None,
             page_offset: int = None,
             page_size: int = None
             ) -> entities.PagedEntities:
        """
        List items in a dataset.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param int page_offset: start page
        :param int page_size: page size
        :return: Pages object
        :rtype: dtlpy.entities.paged_entities.PagedEntities

        **Example**:

        .. code-block:: python

            dataset.items.list(page_offset=0, page_size=100)
        """
        # default filters
        if filters is None:
            filters = entities.Filters()
            filters._user_query = 'false'
        # assert type filters
        elif not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException(error='400',
                                               message='Unknown filters type: {!r}'.format(type(filters)))
        if filters.resource != entities.FiltersResource.ITEM and filters.resource != entities.FiltersResource.ANNOTATION:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.ITEM. Got: {!r}'.format(filters.resource))

        # page size
        if page_size is not None:
            filters.page_size = page_size

        # page offset
        if page_offset is not None:
            filters.page = page_offset

        if filters.resource == entities.FiltersResource.ITEM:
            items_repository = self
        else:
            items_repository = repositories.Annotations(client_api=self._client_api,
                                                        dataset=self._dataset)

        paged = entities.PagedEntities(items_repository=items_repository,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged

    @_api_reference.add(path='/items/{id}', method='get')
    def get(self,
            filepath: str = None,
            item_id: str = None,
            fetch: bool = None,
            is_dir: bool = False
            ) -> entities.Item:
        """
        Get Item object

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str filepath: optional - search by remote path
        :param str item_id: optional - search by id
        :param bool fetch: optional - fetch entity from platform, default taken from cookie
        :param bool is_dir: True if you want to get an item from dir type
        :return: Item object
        :rtype: dtlpy.entities.item.Item

        **Example**:

        .. code-block:: python

            dataset.items.get(item_id='item_id')
        """
        if fetch is None:
            fetch = self._client_api.fetch_entities

        if fetch:
            if item_id is not None:
                success, response = self._client_api.gen_request(req_type="get",
                                                                 path="/items/{}".format(item_id))
                if success:
                    item = self.items_entity.from_json(client_api=self._client_api,
                                                       _json=response.json(),
                                                       dataset=self._dataset,
                                                       project=self._project)
                    # verify input filepath is same as the given id
                    if filepath is not None and item.filename != filepath:
                        logger.warning(
                            "Mismatch found in items.get: filepath is different then item.filename: "
                            "{!r} != {!r}".format(
                                filepath,
                                item.filename))
                else:
                    raise exceptions.PlatformException(response)
            elif filepath is not None:
                filters = entities.Filters()
                filters.pop(field='hidden')
                if is_dir:
                    filters.add(field='type', values='dir')
                filters.recursive = False
                filters.add(field='filename', values=filepath)
                paged_entity = self.list(filters=filters)
                if len(paged_entity.items) == 0:
                    raise exceptions.PlatformException(error='404',
                                                       message='Item not found. filepath= "{}"'.format(filepath))
                elif len(paged_entity.items) > 1:
                    raise exceptions.PlatformException(
                        error='404',
                        message='More than one item found. Please "get" by id. filepath: "{}"'.format(filepath))
                else:
                    item = paged_entity.items[0]
            else:
                raise exceptions.PlatformException(error="400",
                                                   message='Must choose by at least one. "filename" or "item_id"')
        else:
            item = entities.Item.from_json(_json={'id': item_id,
                                                  'filename': filepath},
                                           client_api=self._client_api,
                                           dataset=self._dataset,
                                           is_fetched=False,
                                           project=self._project)
        assert isinstance(item, entities.Item)
        return item

    @_api_reference.add(path='/items/{id}/clone', method='post')
    def clone(self,
              item_id: str,
              dst_dataset_id: str,
              remote_filepath: str = None,
              metadata: dict = None,
              with_annotations: bool = True,
              with_metadata: bool = True,
              with_task_annotations_status: bool = False,
              allow_many: bool = False,
              wait: bool = True):
        """
        Clone item. Read more about cloning datatsets and items in our `documentation <https://dataloop.ai/docs/clone-merge-dataset#cloned-dataset>`_ and `SDK documentation <https://developers.dataloop.ai/tutorials/data_management/data_versioning/chapter/>`_.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str item_id: item to clone
        :param str dst_dataset_id: destination dataset id
        :param str remote_filepath: complete filepath
        :param dict metadata: new metadata to add
        :param bool with_annotations: clone annotations
        :param bool with_metadata: clone metadata
        :param bool with_task_annotations_status: clone task annotations status
        :param bool allow_many: `bool` if True, using multiple clones in single dataset is allowed, (default=False)
        :param bool wait: wait for the command to finish
        :return: Item object
        :rtype: dtlpy.entities.item.Item

        **Example**:

        .. code-block:: python

            dataset.items.clone(item_id='item_id',
                    dst_dataset_id='dist_dataset_id',
                    with_metadata=True,
                    with_task_annotations_status=False,
                    with_annotations=False)
        """
        if metadata is None:
            metadata = dict()
        payload = {"targetDatasetId": dst_dataset_id,
                   "remoteFileName": remote_filepath,
                   "metadata": metadata,
                   "cloneDatasetParams": {
                       "withItemsAnnotations": with_annotations,
                       "withMetadata": with_metadata,
                       "withTaskAnnotationsStatus": with_task_annotations_status},
                   "allowMany": allow_many
                   }
        success, response = self._client_api.gen_request(req_type="post",
                                                         path="/items/{}/clone".format(item_id),
                                                         json_req=payload)
        # check response
        if not success:
            raise exceptions.PlatformException(response)

        command = entities.Command.from_json(_json=response.json(),
                                             client_api=self._client_api)
        if not wait:
            return command
        command = command.wait()

        if 'returnedModelId' not in command.spec:
            raise exceptions.PlatformException(error='400',
                                               message="returnedModelId key is missing in command response: {}"
                                               .format(response))
        cloned_item = self.get(item_id=command.spec['returnedModelId'][0])
        return cloned_item

    @_api_reference.add(path='/items/{id}', method='delete')
    def delete(self,
               filename: str = None,
               item_id: str = None,
               filters: entities.Filters = None):
        """
        Delete item from platform.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        You must provide at least ONE of the following params: item id, filename, filters.

        :param str filename: optional - search item by remote path
        :param str item_id: optional - search item by id
        :param dtlpy.entities.filters.Filters filters: optional - delete items by filter
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            dataset.items.delete(item_id='item_id')
        """
        if item_id is not None:
            success, response = self._client_api.gen_request(req_type="delete",
                                                             path="/items/{}".format(item_id),
                                                             )
        elif filename is not None:
            if not filename.startswith("/"):
                filename = "/" + filename
            items = self.get(filepath=filename)
            if not isinstance(items, list):
                items = [items]
            if len(items) == 0:
                raise exceptions.PlatformException("404", "Item not found")
            elif len(items) > 1:
                raise exceptions.PlatformException(error="404", message="More the 1 item exist by the name provided")
            else:
                item_id = items[0].id
                success, response = self._client_api.gen_request(req_type="delete",
                                                                 path="/items/{}".format(item_id))
        elif filters is not None:
            # prepare request
            success, response = self._client_api.gen_request(req_type="POST",
                                                             path="/datasets/{}/query".format(self.dataset.id),
                                                             json_req=filters.prepare(operation='delete'))
        else:
            raise exceptions.PlatformException("400", "Must provide item id, filename or filters")

        # check response
        if success:
            logger.debug("Item/s deleted successfully")
            return success
        else:
            raise exceptions.PlatformException(response)

    @_api_reference.add(path='/items/{id}', method='patch')
    def update(self,
               item: entities.Item = None,
               filters: entities.Filters = None,
               update_values=None,
               system_update_values=None,
               system_metadata: bool = False):
        """
        Update item metadata.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        You must provide at least ONE of the following params: update_values, system_update_values.

        :param dtlpy.entities.item.Item item: Item object
        :param dtlpy.entities.filters.Filters filters: optional update filtered items by given filter
        :param update_values: optional field to be updated and new values
        :param system_update_values: values in system metadata to be updated
        :param bool system_metadata: True, if you want to update the metadata system
        :return: Item object
        :rtype: dtlpy.entities.item.Item

        **Example**:

        .. code-block:: python

            dataset.items.update(item='item_entity')
        """
        ref = filters is not None and (filters._ref_task or filters._ref_assignment)

        if system_update_values and not system_metadata:
            logger.warning('system metadata will not be updated because param system_metadata is False')

        # check params
        if item is None and filters is None:
            raise exceptions.PlatformException('400', 'must provide either item or filters')

        value_to_update = update_values or system_update_values

        if item is None and not ref and not value_to_update:
            raise exceptions.PlatformException('400',
                                               'Must provide update_values or system_update_values')

        if item is not None and value_to_update:
            raise exceptions.PlatformException('400',
                                               'Cannot provide "update_values" or "system_update_values" with a specific "item" for an individual update. '
                                               'These parameters are intended only for bulk updates using filters.')

        # update item
        if item is not None:
            json_req = miscellaneous.DictDiffer.diff(origin=item._platform_dict,
                                                     modified=item.to_json())
            if not json_req:
                return item
            url_path = "/items/{}".format(item.id)
            if system_metadata:
                url_path += "?system=true"
            success, response = self._client_api.gen_request(req_type="patch",
                                                             path=url_path,
                                                             json_req=json_req)
            if success:
                logger.debug("Item was updated successfully. Item id: {}".format(item.id))
                return self.items_entity.from_json(client_api=self._client_api,
                                                   _json=response.json(),
                                                   dataset=self._dataset)
            else:
                logger.error("Error while updating item")
                raise exceptions.PlatformException(response)
        # update by filters
        else:
            # prepare request
            prepared_filter = filters.prepare(operation='update',
                                              system_update=system_update_values,
                                              system_metadata=system_metadata,
                                              update=update_values)
            success, response = self._client_api.gen_request(req_type="POST",
                                                             path="/datasets/{}/query".format(self.dataset.id),
                                                             json_req=prepared_filter)
            if not success:
                raise exceptions.PlatformException(response)
            else:
                logger.debug("Items were updated successfully.")
                return response.json()

    def download(
            self,
            filters: entities.Filters = None,
            items=None,
            # download options
            local_path: str = None,
            file_types: list = None,
            save_locally: bool = True,
            to_array: bool = False,
            annotation_options: entities.ViewAnnotationOptions = None,
            annotation_filters: entities.Filters = None,
            overwrite: bool = False,
            to_items_folder: bool = True,
            thickness: int = 1,
            with_text: bool = False,
            without_relative_path=None,
            avoid_unnecessary_annotation_download: bool = False,
            include_annotations_in_output: bool = True,
            export_png_files: bool = False,
            filter_output_annotations: bool = False,
            alpha: float = 1,
            export_version=entities.ExportVersion.V1,
            dataset_lock: bool = False,
            lock_timeout_sec: int = None,
            export_summary: bool = False,
            raise_on_error: bool = False,
    ):
        """
        Download dataset items by filters.

        Filters the dataset for items and saves them locally.

        Optional -- download annotation, mask, instance, and image mask of the item.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param List[dtlpy.entities.item.Item] or dtlpy.entities.item.Item items: download Item entity or item_id (or a list of item)
        :param str local_path: local folder or filename to save to.
        :param list file_types: a list of file type to download. e.g ['video/webm', 'video/mp4', 'image/jpeg', 'image/png']
        :param bool save_locally: bool. save to disk or return a buffer
        :param bool to_array: returns Ndarray when True and local_path = False
        :param list annotation_options: download annotations options:  list(dl.ViewAnnotationOptions)
        :param dtlpy.entities.filters.Filters annotation_filters: Filters entity to filter annotations for download
        :param bool overwrite: optional - default = False
        :param bool dataset_lock: optional - default = False
        :param bool export_summary: optional - default = False
        :param int lock_timeout_sec: optional
        :param bool to_items_folder: Create 'items' folder and download items to it
        :param int thickness: optional - line thickness, if -1 annotation will be filled, default =1
        :param bool with_text: optional - add text to annotations, default = False
        :param bool without_relative_path: bool - download items without the relative path from platform
        :param bool avoid_unnecessary_annotation_download: default - False
        :param bool include_annotations_in_output: default - False , if export should contain annotations
        :param bool export_png_files: default - if True, semantic annotations should be exported as png files
        :param bool filter_output_annotations: default - False, given an export by filter - determine if to filter out annotations
        :param float alpha: opacity value [0 1], default 1
        :param str export_version:  exported items will have original extension in filename, `V1` - no original extension in filenames
        :param bool raise_on_error: raise an exception if an error occurs
        :return: generator of local_path per each downloaded item
        :rtype: generator or single item

        **Example**:

        .. code-block:: python

            dataset.items.download(local_path='local_path',
                                 annotation_options=dl.ViewAnnotationOptions,
                                 overwrite=False,
                                 thickness=1,
                                 with_text=False,
                                 alpha=1,
                                 save_locally=True
                                 )
        """
        downloader = repositories.Downloader(self)
        return downloader.download(
            filters=filters,
            items=items,
            local_path=local_path,
            file_types=file_types,
            save_locally=save_locally,
            to_array=to_array,
            annotation_options=annotation_options,
            annotation_filters=annotation_filters,
            overwrite=overwrite,
            to_items_folder=to_items_folder,
            thickness=thickness,
            alpha=alpha,
            with_text=with_text,
            without_relative_path=without_relative_path,
            avoid_unnecessary_annotation_download=avoid_unnecessary_annotation_download,
            include_annotations_in_output=include_annotations_in_output,
            export_png_files=export_png_files,
            filter_output_annotations=filter_output_annotations,
            export_version=export_version,
            dataset_lock=dataset_lock,
            lock_timeout_sec=lock_timeout_sec,
            export_summary=export_summary,
            raise_on_error=raise_on_error
        )

    def upload(
            self,
            # what to upload
            local_path: str,
            local_annotations_path: str = None,
            # upload options
            remote_path: str = "/",
            remote_name: str = None,
            file_types: list = None,
            overwrite: bool = False,
            item_metadata: dict = None,
            output_entity=entities.Item,
            no_output: bool = False,
            export_version: str = entities.ExportVersion.V1,
            item_description: str = None,
            raise_on_error: bool = False,
            return_as_list: bool = False
    ):
        """
        Upload local file to dataset.
        Local filesystem will remain unchanged.
        If "*" at the end of local_path (e.g. "/images/*") items will be uploaded without the head directory.

        **Prerequisites**: Any user can upload items.

        :param str local_path: list of local file, local folder, BufferIO, numpy.ndarray or url to upload
        :param str local_annotations_path: path to dataloop format annotations json files.
        :param str remote_path: remote path to save.
        :param str remote_name: remote base name to save. when upload numpy.ndarray as local path, remote_name with .jpg or .png ext is mandatory
        :param list file_types: list of file type to upload. e.g ['.jpg', '.png']. default is all
        :param dict item_metadata: metadata dict to upload to item or ExportMetadata option to export metadata from annotation file
        :param bool overwrite: optional - default = False
        :param output_entity: output type
        :param bool no_output: do not return the items after upload
        :param str export_version:  exported items will have original extension in filename, `V1` - no original extension in filenames
        :param str item_description: add a string description to the uploaded item
        :param bool raise_on_error: raise an exception if an error occurs
        :param bool return_as_list: return a list of items instead of a generator
        
        :return: Output (generator/single item)
        :rtype: generator or single item

        **Example**:

        .. code-block:: python

            dataset.items.upload(local_path='local_path',
                                 local_annotations_path='local_annotations_path',
                                 overwrite=True,
                                 item_metadata={'Hellow': 'Word'}
                                 )
        """
        # initiate and use uploader
        uploader = repositories.Uploader(items_repository=self, output_entity=output_entity, no_output=no_output)
        return uploader.upload(
            local_path=local_path,
            local_annotations_path=local_annotations_path,
            # upload options
            remote_path=remote_path,
            remote_name=remote_name,
            file_types=file_types,
            # config
            overwrite=overwrite,
            # metadata to upload with items
            item_metadata=item_metadata,
            export_version=export_version,
            item_description=item_description,
            raise_on_error=raise_on_error,
            return_as_list=return_as_list
        )

    @property
    def platform_url(self):
        return self._client_api._get_resource_url(
            "projects/{}/datasets/{}/items".format(self.dataset.project.id, self.dataset.id))

    def open_in_web(self, filepath=None, item_id=None, item=None):
        """
        Open the item in web platform

        **Prerequisites**: You must be in the role of an *owner* or *developer* or be an *annotation manager*/*annotator* with access to that item through task.

        :param str filepath: item file path
        :param str item_id: item id
        :param dtlpy.entities.item.Item item: item entity

        **Example**:

        .. code-block:: python

            dataset.items.open_in_web(item_id='item_id')

        """
        if filepath is not None:
            item = self.get(filepath=filepath)
        if item is not None:
            item.open_in_web()
        elif item_id is not None:
            self._client_api._open_in_web(url=self.platform_url + '/' + str(item_id))
        else:
            self._client_api._open_in_web(url=self.platform_url)

    def update_status(self,
                      status: entities.ItemStatus,
                      items=None,
                      item_ids=None,
                      filters=None,
                      dataset=None,
                      clear=False):
        """
        Update item status in task

        **Prerequisites**: You must be in the role of an *owner* or *developer* or *annotation manager* who has been assigned a task with the item.

        You must provide at least ONE of the following params: items, item_ids, filters.

        :param str status: ItemStatus.COMPLETED, ItemStatus.APPROVED, ItemStatus.DISCARDED
        :param list items: list of items
        :param list item_ids: list of items id
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param dtlpy.entities.dataset.Dataset dataset: dataset object
        :param bool clear: to delete status

        **Example**:

        .. code-block:: python

            dataset.items.update_status(item_ids='item_id', status=dl.ItemStatus.COMPLETED)

        """
        if items is None and item_ids is None and filters is None:
            raise exceptions.PlatformException('400', 'Must provide either items, item_ids or filters')

        if self._dataset is None and dataset is None:
            raise exceptions.PlatformException('400', 'Please provide dataset')
        elif dataset is None:
            dataset = self._dataset

        if filters is not None:
            items = dataset.items.list(filters=filters)
            item_count = items.items_count
        elif items is not None:
            if isinstance(items, entities.PagedEntities):
                item_count = items.items_count
            else:
                if not isinstance(items, list):
                    items = [items]
                item_count = len(items)
                items = [items]
        else:
            if not isinstance(item_ids, list):
                item_ids = [item_ids]
            item_count = len(item_ids)
            items = [[dataset.items.get(item_id=item_id, fetch=False) for item_id in item_ids]]

        pool = self._client_api.thread_pools(pool_name='item.status_update')
        jobs = [None for _ in range(item_count)]
        # call multiprocess wrapper to run service on each item in list
        for page in items:
            for i_item, item in enumerate(page):
                jobs[i_item] = pool.submit(item.update_status,
                                           **{'status': status,
                                              'clear': clear})

        # get all results
        results = [j.result() for j in jobs]
        out_success = [r for r in results if r is True]
        out_errors = [r for r in results if r is False]
        if len(out_errors) == 0:
            logger.debug('Item/s updated successfully. {}/{}'.format(len(out_success), len(results)))
        else:
            logger.error(out_errors)
            logger.error('Item/s updated with {} errors'.format(len(out_errors)))

    def make_dir(self, directory, dataset: entities.Dataset = None) -> entities.Item:
        """
        Create a directory in a dataset.

        **Prerequisites**: All users.

        :param str directory: name of directory
        :param dtlpy.entities.dataset.Dataset dataset: dataset object
        :return: Item object
        :rtype: dtlpy.entities.item.Item

        **Example**:

        .. code-block:: python

            dataset.items.make_dir(directory='directory_name')
        """
        if self._dataset_id is None and dataset is None:
            raise exceptions.PlatformException('400', 'Please provide parameter dataset')

        payload = {
            'type': 'dir',
            'path': directory
        }
        headers = {'content-type': 'application/x-www-form-urlencoded'}
        success, response = self._client_api.gen_request(req_type="post",
                                                         headers=headers,
                                                         path="/datasets/{}/items".format(self._dataset_id),
                                                         data=payload)
        if success:
            item = self.items_entity.from_json(client_api=self._client_api,
                                               _json=response.json(),
                                               dataset=self._dataset)
        else:
            raise exceptions.PlatformException(response)

        return item

    def move_items(self,
                   destination: str,
                   filters: entities.Filters = None,
                   items=None,
                   dataset: entities.Dataset = None
                   ) -> bool:
        """
        Move items to another directory.
        If directory does not exist we will create it

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str destination: destination directory
        :param dtlpy.entities.filters.Filters filters: optional - either this or items. Query of items to move
        :param items: optional - either this or filters. A list of items to move
        :param dtlpy.entities.dataset.Dataset dataset: dataset object
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            dataset.items.move_items(destination='directory_name')
        """
        if filters is None and items is None:
            raise exceptions.PlatformException('400', 'Must provide either filters or items')

        dest_dir_filter = entities.Filters(resource=entities.FiltersResource.ITEM, field='type', values='dir')
        dest_dir_filter.recursive = False
        dest_dir_filter.add(field='filename', values=destination)
        dirs_page = self.list(filters=dest_dir_filter)

        if dirs_page.items_count == 0:
            directory = self.make_dir(directory=destination, dataset=dataset)
        elif dirs_page.items_count == 1:
            directory = dirs_page.items[0]
        else:
            raise exceptions.PlatformException('404', 'More than one directory by the name of: {}'.format(destination))

        if filters is not None:
            items = self.list(filters=filters)
        elif isinstance(items, list):
            items = [items]
        elif not isinstance(items, entities.PagedEntities):
            raise exceptions.PlatformException('400', 'items must be a list of items or a pages entity not {}'.format(
                type(items)))

        item_ids = list()
        for page in items:
            for item in page:
                item_ids.append(item.id)

        success, response = self._client_api.gen_request(req_type="put",
                                                         path="/datasets/{}/items/{}".format(self._dataset_id,
                                                                                             directory.id),
                                                         json_req=item_ids)
        if not success:
            raise exceptions.PlatformException(response)

        return success
    
    def task_scores(self, item_id: str, task_id: str, page_offset: int = 0, page_size: int = 100):
        """
        Get item score

        **Prerequisites**: You must be able to read the task

        :param str item_id: item id
        :param str task_id: task id
        :param int page_offset: start page
        :param int page_size: page size
        :return: page of item scores

        **Example**:

        .. code-block:: python
        
            dataset.items.item_score(item_id='item_id', task_id='task_id')

        """

        if item_id is None:
            raise exceptions.PlatformException('400', 'Must provide item id')

        if task_id is None:
            raise exceptions.PlatformException('400', 'Must provide task id')

        success, response = self._client_api.gen_request(req_type="get",
                                                         path="/scores/tasks/{}/items/{}?page={}&pageSize={}"
                                                         .format(task_id, item_id, page_offset, page_size))
        if success:
            return response.json()
        else:
            raise exceptions.PlatformException(response)




import subprocess
import logging
import os

logger = logging.getLogger(name='dtlpy')


class GitUtils:
    """
    Performs git related methods
    """

    @staticmethod
    def git_included(path):
        """
        Get only included git repo files based on .gitignore file

        :param path: directory - str
        :return: list()
        """
        included_files = list()

        try:
            p = subprocess.Popen(['git', '--git-dir', os.path.join(path, '.git'), 'ls-files'],
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            output, err = p.communicate()
            string_output = str(output, 'utf-8')
            included_files = string_output.split('\n')
        except Exception:
            logging.warning('Error getting git info for git repository in: {}'.format(path))
            # include all files
            for r, d, f in os.walk(path):
                for folder in d:
                    included_files.append(os.path.join(r, folder))

        return included_files

    @staticmethod
    def is_git_repo(path):
        """
        Check if directory is a git repo

        :param path: directory - str
        :return: True/False
        """
        try:
            p = subprocess.Popen(['git', '--git-dir', os.path.join(path, '.git'), '--work-tree', path, 'status'],
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            output, err = p.communicate()
            if p.returncode != 0 and 'not a git repository' in str(err):
                response = False
            elif p.returncode == 0 and 'On branch' in str(output):
                response = True
            else:
                response = False
        except Exception:
            response = False
            logging.warning('Error getting git info for git repository in: {}'.format(path))
        return response

    @staticmethod
    def git_status(path):
        """
        Get git repository git status

        :param path: directory - str
        :return: String
        """
        status = dict()
        try:
            p = subprocess.Popen(['git', '--git-dir', os.path.join(path, '.git'), '--work-tree', path, 'status'],
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            output, err = p.communicate()
            output_lines = str(output, 'utf-8').splitlines()
            branch = output_lines[0].replace('On branch ', '').strip()

            logs = GitUtils.git_log(path)
            status = {'branch': branch,
                      'commit_id': logs[0]['commit'],
                      'commit_author': logs[0]['Author'],
                      'commit_message': logs[0]['message']}
        except Exception:
            logging.warning('Error getting git info for git repository in: {}'.format(path))

        return status

    @staticmethod
    def git_url(path):
        """
        Get git remote url

        :param path: directory - str
        :return: String
        """
        url = ''
        try:
            p = subprocess.Popen(
                ['git', '--git-dir', os.path.join(path, '.git'), 'config', '--get', 'remote.origin.url'],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)
            output, err = p.communicate()
            url = str(output, 'utf-8').splitlines()[0]

        except Exception:
            logging.warning('Error getting git info for git repository in: {}'.format(path))

        return url

    @staticmethod
    def git_log(path):
        """
        Get git repository git log

        :param path: directory - str
        :return: log as list()
        """
        log = list()
        try:
            log_limit = 100
            p = subprocess.Popen(['git', '--git-dir', os.path.join(path, '.git'), '--work-tree', path, 'log'],
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            output, err = p.communicate()
            string_output = str(output, 'utf-8').split('\ncommit')
            for output in string_output:
                output = output.split('\n')
                if output[0].startswith('commit'):
                    output[0] = output[0].replace('commit', '')
                log_line = {
                    'commit': output[0].strip(),
                    'Author': output[1].replace('Author:', '').strip(),
                    'Date': output[2].replace('Date:', '').strip(),
                    'message': output[4].strip(),
                }
                log.append(log_line)
            log = log[0:log_limit]
        except Exception:
            logging.warning('Error getting git log for git repository in: {}'.format(path))

        return log

    @staticmethod
    def git_clone(path, git_url, **kwargs):
        """
        Clone git repo to local path
        :param path: `str` local path to clone to
        :param git_url: `str` git url to clone from
        :return `bool` for successful clone
        """
        cmd = ''
        if not os.path.isdir(path):
            os.makedirs(path)
        try:
            username = kwargs.get('username')
            password = kwargs.get('password')
            if username is not None and password is not None:
                if git_url.startswith('https://'):
                    git_url = git_url.replace('https://', 'https://{}:{}@'.format(username, password))
                elif git_url.startswith('http://'):
                    git_url = git_url.replace('http://', 'http://{}:{}@'.format(username, password))
                else:
                    git_url = 'https://{}:{}@{}'.format(username, password, git_url)
            tag = kwargs.get('tag')
            branch = kwargs.get('branch')
            if tag is not None:
                branch_cmd = ['--branch', tag]
            elif branch is not None:
                branch_cmd = ['--branch', branch]
            else:
                branch_cmd = []

            cmd = ['git', 'clone'] + branch_cmd + [git_url, path]
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            output, err = [std.decode() for std in p.communicate()]
            exit_code = p.returncode
            response = not exit_code
            if exit_code:
                logging.error('Error executing:  {ps1} $ {cmd}\n{err}'.format(ps1=path, cmd=' '.join(cmd), err=err))
        except Exception:
            response = False
            logging.exception('Error cloning git with cmd: {}'.format(cmd))
        return response

    @staticmethod
    def git_command(path, cmd, show=False):
        """
        Execute command for a local git repo in path
        :param path: `str` local path for the git repo
        :param cmd:  `str` or `list` of `str` specifying the command to run
        :param show: `bool` if True, prints the stdout of the process (default=False)
        :return:  `bool` if command was successful or not
        """
        prev_dir = os.getcwd()
        if isinstance(cmd, str):
            cmd = cmd.split()

        try:
            os.chdir(path)
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            output, err = p.communicate()
            output, err = output.decode(), err.decode()
            exit_code = p.returncode
            response = not exit_code
            # Print the process stdout / stderr
            if show:
                print(output)
            if exit_code:
                logging.error('Error executing:  {ps1} $ {cmd}\n{err}'.format(ps1=path, cmd=' '.join(cmd), err=err))
        except Exception:
            response = False
            logging.critical('Error executing:  {ps1} $ {cmd}'.format(ps1=path, cmd=' '.join(cmd)))
        finally:
            os.chdir(prev_dir)
        return response
